# AUTOGENERATED! DO NOT EDIT! File to edit: ../../nbs/data_utils/features.ipynb.

# %% ../../nbs/data_utils/features.ipynb 3
from __future__ import annotations
from .preprocessing import *
from .transforms import *
from ..utils import get_config, gumbel_softmax, load_pytree, save_pytree
from ..import_essentials import *
from copy import deepcopy

# %% auto 0
__all__ = ['Feature', 'FeaturesList']

# %% ../../nbs/data_utils/features.ipynb 5
class Feature:
    """THe feature class which represents a column in the dataset."""
    
    def __init__(
        self,
        name: str,
        data: np.ndarray,
        transformation: str | BaseTransformation | dict,
        transformed_data = None,
        is_immutable: bool = False,
        is_categorical: bool = None,
    ):
        self.name = name
        self._data = data
        self._transformation = self._dispatch_transformation(transformation)
        self._transformed_data = transformed_data
        self._is_immutable = is_immutable
        self._is_categorical = self._init_is_categorical(is_categorical)

    def with_transformed_data(
        self,
        transformed_data: np.ndarray,
        **kwargs        
    ) -> Feature:
        """Create a new feature with transformed data."""
        
        org_data = self.inverse_transform(transformed_data)
        return Feature(
            name=self.name,
            data=org_data,
            transformation=self.transformation,
            transformed_data=transformed_data,
            is_immutable=self.is_immutable,
            is_categorical=self.is_categorical,
        )
    
    @property
    def data(self) -> jax.Array:
        return self._data
    
    @property
    def is_immutable(self) -> bool:
        return self._is_immutable
    
    @property
    def transformation(self) -> BaseTransformation:
        return self._transformation
    
    @property
    def is_categorical(self) -> bool:
        return self._is_categorical

    @property
    def transformed_data(self) -> jax.Array:
        if self._transformed_data is None:
            return self.fit_transform(self.data)
        else:
            return self._transformed_data

    @classmethod
    def from_dict(cls, d):
        return cls(**d)
    
    def to_dict(self):
        return {
            'name': self.name,
            'data': self.data,
            'transformed_data': self.transformed_data,
            'transformation': self.transformation.to_dict(),
            'is_immutable': self.is_immutable,
            'is_categorical': self.is_categorical,
        }
    
    def __repr__(self):
        return "Feature(" + ",\n".join([
            f"{k}={v}" for k, v in self.to_dict().items()]) + ")"
    
    __str__ = __repr__

    def __get_item__(self, idx):
        return self.to_dict().update({
            'data': self.data[idx],
            'transformed_data': self.transformed_data[idx],
        })
    
    def _dispatch_transformation(self, transformation: str | dict | BaseTransformation):
        T = FEATURE_TRANSFORMATIONS
        if isinstance(transformation, str):
            if transformation not in T.keys():
                raise ValueError(f"Unknown transformation: {transformation}")
            return T[transformation]()
        elif isinstance(transformation, dict):
            # TODO: only supported transformation can be used for serialization
            t_name = transformation['name']
            if t_name not in T.keys():
                raise ValueError("Only supported transformation can be inited from dict. "
                                 f"Got {t_name}, but should be one of {T.keys()}.")
            return T[t_name]().from_dict(transformation)
        elif isinstance(transformation, BaseTransformation):
            return deepcopy(transformation)
        else:
            raise ValueError(f"Unknown transformation: {transformation}")
        
    def _init_is_categorical(self, is_categorical: bool = None):
        if is_categorical is None:
            return self.transformation.is_categorical
        else:
            if hasattr(self, '_is_categorical'):
                assert self.is_categorical == is_categorical
            return is_categorical

    def set_transformation(self, transformation: str | dict | BaseTransformation) -> Feature:
        self._transformation = self._dispatch_transformation(transformation)
        self._is_categorical = self._init_is_categorical()
        self._transformed_data = None # Reset transformed data
        return self
    
    def fit(self):
        self.transformation.fit(self.data)
        return self
    
    def transform(self, xs):
        return self.transformation.transform(xs)

    def fit_transform(self, xs):
        return self.transformation.fit_transform(xs)
    
    def inverse_transform(self, xs):
        return self.transformation.inverse_transform(xs)
    
    def apply_constraints(self, xs, cfs, hard: bool = False, rng_key=None, **kwargs):
        return jax.lax.cond(
            self.is_immutable,
            true_fun=lambda xs: jnp.broadcast_to(xs, cfs.shape),
            false_fun=lambda _: self.transformation.apply_constraints(xs, cfs, hard=hard, rng_key=rng_key, **kwargs),
            operand=xs,
        )
    
    def compute_reg_loss(self, xs, cfs, hard: bool = False):
        return self.transformation.compute_reg_loss(xs, cfs, hard)

# %% ../../nbs/data_utils/features.ipynb 8
class FeaturesList:
    def __init__(
        self,
        features: list[Feature] | FeaturesList,
        *args, **kwargs
    ):
        if isinstance(features, FeaturesList):
            self._features = features.features
            self._feature_indices = features.feature_indices
            self._transformed_data = features.transformed_data
        elif isinstance(features, Feature):
            self._features = [features]
        elif isinstance(features, list):
            if len(features) > 0 and not isinstance(features[0], Feature):
                raise ValueError(f"Invalid features type: {type(features[0]).__name__}")
            self._features = features
        else:
            raise ValueError(f"Unknown features type. Got {type(features).__name__}")
        
        # Record the current position of the features
        self.pose = 0

    # Iterator
    def __len__(self):
        return len(self._features)
    
    def __iter__(self):
        return self
    
    def __next__(self):
        if self.pose < len(self):
            feat = self._features[self.pose]
            self.pose += 1
            return feat
        else:
            self.pose = 0
            raise StopIteration
        
    # Indexing
    def __getitem__(self, idx: str | list[str]) -> Feature | list[Feature]:
        if not hasattr(self, "_feature_name_list_indices"):
            self._feature_name_list_indices = {feat.name: i for i, feat in enumerate(self._features)}
        indices = self._feature_name_list_indices
        if isinstance(idx, str):
            if idx not in indices:
                raise ValueError(f"Invalid feature name: {idx}")
            return self._features[indices[idx]]
        elif isinstance(idx, list):
            return [self[i] for i in idx]
        else:
            raise ValueError(f"Invalid idx type: {type(idx).__name__}")
    
    #############################
    # Properties
    #############################
    @property
    def features(self) -> list[Feature]: # Return [Feature(...), ...]
        return self._features

    @property
    def feature_indices(self) -> list[tuple[int, int]]: # Return [(start, end), ...]
        if not hasattr(self, "_feature_indices") or self._feature_indices is None or len(self._feature_indices) == 0:
            self._transform_data()
        return self._feature_indices
    
    @property
    def features_and_indices(self) -> list[tuple[Feature, tuple[int, int]]]: # Return [(Feature(...), (start, end)), ...]
        return list(zip(self.features, self.feature_indices))
    
    @property
    def feature_name_indices(self) -> dict[str, tuple[int, int]]: # Return {feature_name: (feat_idx, start, end), ...}
        if not hasattr(self, "_feature_name_indices") or self._feature_name_indices is None:
            self._transform_data()
        return self._feature_name_indices
    
    @property
    def transformed_data(self) -> jax.Array:
        if not hasattr(self, "_transformed_data") or self._transformed_data is None:
            self._transform_data()
        return self._transformed_data
    
    #############################
    # Methods
    #############################
    def with_transformed_data(
        self,
        transformed_data: np.ndarray,
        **kwargs        
    ) -> FeaturesList:
        """Create a new feature with transformed data."""
        
        return FeaturesList(
            features=[feat.with_transformed_data(transformed_data[:, start:end], **kwargs) 
                      for feat, (start, end) in self.features_and_indices],
        )
    
    def set_transformations(self, feature_names_to_transformation: dict[str, BaseTransformation]) -> FeaturesList:
        """Set the transformations for the features."""
        
        if not isinstance(feature_names_to_transformation, dict):
            raise ValueError(f"Invalid feature_names_to_transformation type: "
                             f"{type(feature_names_to_transformation).__name__}."
                             f"Should be dict[str, BaseTransformation]")
        
        for feat, transformation in feature_names_to_transformation.items():
            self[feat].set_transformation(transformation)
        self._transformed_data = None # Reset transformed data
        return self
        
    def _transform_data(self):
        self._feature_indices = []
        self._feature_name_indices = {}
        self._transformed_data = []
        start, end = 0, 0
        for i, feat in enumerate(self.features):
            transformed_data = feat.transformed_data
            end += transformed_data.shape[-1]
            self._feature_indices.append((start, end))
            self._feature_name_indices[feat.name] = (i, start, end)
            self._transformed_data.append(transformed_data)
            start = end

        self._transformed_data = jnp.concatenate(self._transformed_data, axis=-1)
    
    def transform(self, data: dict[str, jax.Array]):
        if not isinstance(data, dict):
            raise ValueError(f"Invalid data type: {type(data).__name__}, should be dict[str, jax.Array]")

        transformed_data = [None] * len(self.features)
        for feat_name, val in data.items():
            feat_idx, _, _ = self.feature_name_indices[feat_name]
            feat = self.features[feat_idx]
            transformed_data[feat_idx] = feat.transform(val)
        return jnp.concatenate(transformed_data, axis=-1)

    def inverse_transform(self, xs) -> dict[str, jax.Array]:
        orignial_data = {}
        for feat, (start, end) in self.features_and_indices:
            orignial_data[feat.name] = feat.inverse_transform(xs[:, start:end])
        return orignial_data

    def apply_constraints(self, xs, cfs, hard: bool = False, rng_key=None, **kwargs):
        return jnp.concatenate(
            [feat.apply_constraints(xs[:, start:end], cfs[:, start:end], 
                                    hard=hard, rng_key=rng_key, **kwargs) for feat, (start, end) in self.features_and_indices], axis=-1)
    
    def compute_reg_loss(self, xs, cfs, hard: bool = False):
        reg_loss = 0.
        for feat, (start, end) in self.features_and_indices:
            reg_loss += feat.compute_reg_loss(xs[:, start:end], cfs[:, start:end], hard)
        return reg_loss

    def to_dict(self):
        return {
            'features': [feat.to_dict() for feat in self.features],
        }
    
    @classmethod
    def from_dict(cls, d):
        return cls(
            features=[Feature.from_dict(feat) for feat in d['features']],
        )
    
    def to_pandas(self, use_transformed: bool = False) -> pd.DataFrame:
        if use_transformed:
            # data = {feat.name: feat.transformed_data.reshape(-1) for feat in self.features}
            raise NotImplementedError
        else:
            data = {feat.name: feat.data.reshape(-1) for feat in self.features}
        return pd.DataFrame(data=data)

    def save(self, saved_dir):
        os.makedirs(saved_dir, exist_ok=True)
        save_pytree(self.to_dict(), saved_dir)
        
    @classmethod
    def load_from_path(cls, saved_dir):
        return cls.from_dict(load_pytree(saved_dir))



"""Functions for training models"""

# AUTOGENERATED! DO NOT EDIT! File to edit: ../../nbs/legacy/04_trainer.ipynb.

# %% ../../nbs/legacy/04_trainer.ipynb 3
from __future__ import annotations
from .import_essentials import *
from .module import BaseTrainingModule, PredictiveTrainingModule
from .logger import Logger
from .utils import validate_configs, load_json
from .ckpt_manager import CheckpointManager, load_checkpoint
from ..data_module import DataModule
from urllib.request import urlretrieve
from keras.src.trainers.epoch_iterator import EpochIterator


# %% auto 0
__all__ = ['TrainingConfigs', 'train_model_with_states', 'train_model']

# %% ../../nbs/legacy/04_trainer.ipynb 4
class TrainingConfigs(BaseParser):
    """Configurator of `train_model`."""
    
    n_epochs: int = Field(
        description="Number of epochs."
    )
    batch_size: int = Field(
        description="Batch size."
    )
    monitor_metrics: Optional[str] = Field(
        None, description="Monitor metrics used to evaluate the training result after each epoch."
    )
    seed: int = Field(
        42, description="Seed for generating random number."
    )
    log_dir: str = Field(
        "log", description="The name for the directory that holds logged data during training."
    )
    logger_name: str = Field(
        "debug", description="The name for the directory that holds logged data during training under log directory."
    )
    log_on_step: bool = Field(
        False, description="Log the evaluate metrics at the current step."
    )
    max_n_checkpoints: int = Field(
        3, description="Maximum number of checkpoints stored."
    )

    @property
    def PRNGSequence(self):
        return hk.PRNGSequence(self.seed)


# %% ../../nbs/legacy/04_trainer.ipynb 5
def train_model_with_states(
    training_module: BaseTrainingModule,
    params: hk.Params,
    opt_state: optax.OptState,
    data_module: DataModule,
    t_configs: Dict[str, Any] | TrainingConfigs,
) -> Tuple[hk.Params, optax.OptState]:
    """Train models with `params` and `opt_state`."""

    t_configs = validate_configs(t_configs, TrainingConfigs)
    keys = t_configs.PRNGSequence
    # define logger
    logger = Logger(
        log_dir=t_configs.log_dir,
        name=t_configs.logger_name,
        on_step=t_configs.log_on_step,
    )
    logger.save_hyperparams(t_configs.dict())
    if hasattr(training_module, "hparams") and training_module.hparams is not None:
        logger.save_hyperparams(training_module.hparams)

    training_module.init_logger(logger)
    # define checkpoint manageer
    if t_configs.monitor_metrics is None:
        monitor_metrics = None
    else:
        monitor_metrics = f"{t_configs.monitor_metrics}_epoch"

    ckpt_manager = CheckpointManager(
        log_dir=Path(training_module.logger.log_dir) / "checkpoints",
        monitor_metrics=monitor_metrics,
        max_n_checkpoints=t_configs.max_n_checkpoints,
    )
    # dataloaders
    # train_loader = jdl.DataLoader(jdl.ArrayDataset(*data_module['train']), backend='jax', batch_size=t_configs.batch_size, shuffle=True) 
    epoch_iterator = EpochIterator(*data_module['train'], batch_size=t_configs.batch_size, shuffle=True)
    val_epoch_iterator = EpochIterator(*data_module['test'], batch_size=t_configs.batch_size, shuffle=False)

    @jax.jit
    def train_step(params, opt_state, key, batch):
        return training_module.training_step(params, opt_state, key, batch)
    
    # start training
    for epoch in range(t_configs.n_epochs):
        training_module.logger.on_epoch_started()
        # for step, batch in epoch_iterator.enumerate_epoch('np'):
        with tqdm(
            epoch_iterator.enumerate_epoch(), 
            unit="batch", 
            leave=epoch == t_configs.n_epochs - 1,
            total=epoch_iterator.num_batches
        ) as t_loader:
            t_loader.set_description(f"Epoch {epoch}")
            for step, batch in t_loader:
                x, y = batch[0]
                logs, (params, opt_state) = train_step(params, opt_state, next(keys), (x, y))
                # TODO: tqdm becomes the bottleneck
                t_loader.set_postfix(**logs)
        
        # validation
        for step, batch in val_epoch_iterator.enumerate_epoch():
            x, y = batch[0]
            logs = training_module.validation_step(params, next(keys), (x, y))
            # logger.log(logs)
        epoch_logs = training_module.logger.on_epoch_finished()
        ckpt_manager.update_checkpoints(params, opt_state, epoch_logs, epoch)

    training_module.logger.close()
    return params, opt_state


# %% ../../nbs/legacy/04_trainer.ipynb 6
def train_model(
    training_module: BaseTrainingModule, # Training module
    data_module: DataModule, # Data module
    batch_size=128, # Batch size
    epochs=1, # Number of epochs
    **fit_kwargs # Positional arguments for `keras.Model.fit`
) -> Tuple[hk.Params, optax.OptState]:
    """Train models."""
    t_configs = TrainingConfigs(
        n_epochs=epochs,
        batch_size=batch_size,
        **fit_kwargs
    )
    keys = t_configs.PRNGSequence 
    params, opt_state = training_module.init_net_opt(data_module, next(keys))
    return train_model_with_states(
        training_module=training_module,
        params=params,
        opt_state=opt_state,
        data_module=data_module,
        t_configs=t_configs,
    )



# AUTOGENERATED! DO NOT EDIT! File to edit: ../../nbs/methods/00_base.ipynb.

# %% auto 0
__all__ = ['CFModule', 'ParametricCFModule']

# %% ../../nbs/methods/00_base.ipynb 2
from ..import_essentials import *
from ..base import BaseConfig, BaseModule, PredFnMixedin, TrainableMixedin

# %% ../../nbs/methods/00_base.ipynb 3
def default_apply_constraints_fn(x, cf, hard, **kwargs):
    return cf

def default_compute_reg_loss_fn(x, cf, **kwargs):
    return 0.

# %% ../../nbs/methods/00_base.ipynb 4
class CFModule(BaseModule):
    """Base class for all counterfactual modules."""

    def __init__(
        self, 
        config,
        *, 
        name: str = None,
        apply_constraints_fn = None,
        compute_reg_loss_fn = None,
        **kwargs
    ):
        super().__init__(config, name=name)
        self._apply_constraints_fn = apply_constraints_fn
        self._compute_reg_loss_fn = compute_reg_loss_fn
        self.data_module = None

    def set_data_module(self, data_module):
        self.data_module = data_module

    def set_apply_constraints_fn(self, apply_constraints_fn: Callable):
        self._apply_constraints_fn = apply_constraints_fn

    def set_compute_reg_loss_fn(self, compute_reg_loss_fn: Callable):
        self._compute_reg_loss_fn = compute_reg_loss_fn
    
    def apply_constraints(self, *args, **kwargs) -> Array:
        if self._apply_constraints_fn is not None:
            return self._apply_constraints_fn(*args, **kwargs)
        else:
            return default_apply_constraints_fn(*args, **kwargs)
    
    def compute_reg_loss(self, *args, **kwargs):
        if self._compute_reg_loss_fn is not None:
            return self._compute_reg_loss_fn(*args, **kwargs)
        else:
            return default_compute_reg_loss_fn(*args, **kwargs)

    def before_generate_cf(self, *args, **kwargs):
        pass

    def generate_cf(
        self,
        x: Array,
        pred_fn: Callable = None,
        y_target: Array = None,
        rng_key: jrand.PRNGKey = None,
        **kwargs
    ) -> Array: # Return counterfactual of x.
        raise NotImplementedError
    
    __ALL__ = [
        "set_apply_constraints_fn",
        "set_compute_reg_loss_fn",
        "apply_constraints",
        "compute_reg_loss",
        "save",
        "load_from_path",
        "before_generate_cf",
        "generate_cf"
    ]

# %% ../../nbs/methods/00_base.ipynb 5
class ParametricCFModule(CFModule, TrainableMixedin):
    """Base class for parametric counterfactual modules."""
    
    def train(self, data, pred_fn, **kwargs):
        """Train the module."""
        raise NotImplementedError



# AUTOGENERATED! DO NOT EDIT! File to edit: ../../nbs/data_utils/preprocessing.ipynb.

# %% ../../nbs/data_utils/preprocessing.ipynb 3
from __future__ import annotations
import numpy as np
import einops

# %% auto 0
__all__ = ['DataPreprocessor', 'MinMaxScaler', 'EncoderPreprocessor', 'OrdinalPreprocessor', 'OneHotEncoder']

# %% ../../nbs/data_utils/preprocessing.ipynb 5
def _check_xs(xs: np.ndarray, name: str):
    if xs.ndim > 2 or (xs.ndim == 2 and xs.shape[1] != 1):
        raise ValueError(f"`{name}` only supports array with a single feature, but got shape={xs.shape}.")

# %% ../../nbs/data_utils/preprocessing.ipynb 6
class DataPreprocessor:

    def __init__(
        self, 
        name: str = None # The name of the preprocessor. If None, the class name will be used.
    ):
        """Base class for data preprocessors."""
        self.name = name or self.__class__.__name__
    
    def fit(self, xs, y=None):
        """Fit the preprocessor with `xs` and `y`."""
        raise NotImplementedError
    
    def transform(self, xs):
        """Transform `xs`."""
        raise NotImplementedError
    
    def fit_transform(self, xs, y=None):
        """Fit the preprocessor with `xs` and `y`, then transform `xs`."""
        self.fit(xs, y)
        return self.transform(xs)
    
    def inverse_transform(self, xs):
        """Inverse transform `xs`."""
        raise NotImplementedError
    
    def to_dict(self) -> dict:
        """Convert the preprocessor to a dictionary."""
        raise NotImplementedError
    
    def from_dict(self, params: dict):
        """Load the attributes of the preprocessor from a dictionary."""
        raise NotImplementedError
        
    __ALL__ = ["fit", "transform", "fit_transform", "inverse_transform", "to_dict", "from_dict"]

# %% ../../nbs/data_utils/preprocessing.ipynb 7
class MinMaxScaler(DataPreprocessor): 
    def __init__(self):
        super().__init__(name="minmax")
        
    def fit(self, xs, y=None):
        _check_xs(xs, name="MinMaxScaler")
        self.min_ = xs.min(axis=0)
        self.max_ = xs.max(axis=0)
        return self
    
    def transform(self, xs):
        return (xs - self.min_) / (self.max_ - self.min_)
    
    def inverse_transform(self, xs):
        return xs * (self.max_ - self.min_) + self.min_
    
    def from_dict(self, params: dict):
        self.min_ = params["min_"]
        self.max_ = params["max_"]
        return self
    
    def to_dict(self) -> dict:
        return {"min_": self.min_, "max_": self.max_}

# %% ../../nbs/data_utils/preprocessing.ipynb 13
def _unique(xs):
    if xs.dtype == object:
        # Note: np.unique does not work with object dtype
        # We will enforce xs to be string type
        # It assumes that xs is a list of strings, and might not work
        # for other cases (e.g., list of string and numbers)
        return np.unique(xs.astype(str))
    return np.unique(xs)

# %% ../../nbs/data_utils/preprocessing.ipynb 14
class EncoderPreprocessor(DataPreprocessor):
    """Encode categorical features as an integer array."""
    def _fit(self, xs, y=None):
        _check_xs(xs, name="EncoderPreprocessor")
        self.categories_ = _unique(xs)

    def _transform(self, xs):
        """Transform data to ordinal encoding."""
        if xs.dtype == object:
            xs = xs.astype(str)
        ordinal = np.searchsorted(self.categories_, xs)
        # return einops.rearrange(ordinal, 'k n -> n k')
        return ordinal
    
    def _inverse_transform(self, xs):
        """Transform ordinal encoded data back to original data."""
        return self.categories_[xs.T].T
    
    def from_dict(self, params: dict):
        self.categories_ = params["categories_"]
        return self
    
    def to_dict(self) -> dict:
        return {"categories_": self.categories_}

# %% ../../nbs/data_utils/preprocessing.ipynb 15
class OrdinalPreprocessor(EncoderPreprocessor):
    """Ordinal encoder for a single feature."""
    
    def fit(self, xs, y=None):
        self._fit(xs, y)
        return self
    
    def transform(self, xs):
        if xs.ndim == 1:
            raise ValueError(f"OrdinalPreprocessor only supports 2D array with a single feature, "
                             f"but got shape={xs.shape}.")
        return self._transform(xs)
    
    def inverse_transform(self, xs):
        return self._inverse_transform(xs)

# %% ../../nbs/data_utils/preprocessing.ipynb 17
class OneHotEncoder(EncoderPreprocessor):
    """One-hot encoder for a single categorical feature."""
    
    def fit(self, xs, y=None):
        self._fit(xs, y)
        return self

    def transform(self, xs):
        if xs.ndim == 1:
            raise ValueError(f"OneHotEncoder only supports 2D array with a single feature, "
                             f"but got shape={xs.shape}.")
        xs_int = self._transform(xs)
        one_hot_feats = np.eye(len(self.categories_))[xs_int]
        return einops.rearrange(one_hot_feats, 'n k d -> n (k d)')

    def inverse_transform(self, xs):
        xs_int = np.argmax(xs, axis=-1)
        return self._inverse_transform(xs_int).reshape(-1, 1)



# AUTOGENERATED! DO NOT EDIT! File to edit: ../../nbs/data_utils/transform.ipynb.

# %% ../../nbs/data_utils/transform.ipynb 3
from __future__ import annotations
from .preprocessing import *
from ..utils import get_config, gumbel_softmax
from ..import_essentials import *

# %% auto 0
__all__ = ['FEATURE_TRANSFORMATIONS', 'BaseTransformation', 'MinMaxTransformation', 'SoftmaxTransformation',
           'GumbelSoftmaxTransformation', 'OneHotTransformation', 'OrdinalTransformation', 'IdentityTransformation']

# %% ../../nbs/data_utils/transform.ipynb 5
class BaseTransformation:
    """Base class for all transformations."""
    
    def __init__(self, name, transformer: DataPreprocessor = None):
        self.name = name
        self.transformer = transformer

    @property
    def is_categorical(self) -> bool:   raise NotImplementedError

    def fit(self, xs, y=None):          raise NotImplementedError

    def transform(self, xs):            raise NotImplementedError

    def fit_transform(self, xs, y=None):raise NotImplementedError

    def inverse_transform(self, xs):    raise NotImplementedError

    def apply_constraints(self, xs, cfs, hard, rng_key, **kwargs):
        raise NotImplementedError
    
    def compute_reg_loss(self, xs, cfs, hard: bool = False):
        raise NotImplementedError

    def from_dict(self, params):        raise NotImplementedError    

    def to_dict(self):                  raise NotImplementedError

# %% ../../nbs/data_utils/transform.ipynb 6
class _DefaultTransformation(BaseTransformation):

    @property
    def is_categorical(self) -> bool:
        if self.transformer is None:
            return False
        return isinstance(self.transformer, EncoderPreprocessor)

    def fit(self, xs, y=None):
        if self.transformer is not None:
            self.transformer.fit(xs)
        return self
    
    def transform(self, xs):
        if self.transformer is None:
            return xs
        return self.transformer.transform(xs)

    def fit_transform(self, xs, y=None):
        if self.transformer is None:
            return xs
        return self.transformer.fit_transform(xs)
    
    def inverse_transform(self, xs):
        if self.transformer is None:
            return xs
        return self.transformer.inverse_transform(xs)

    def apply_constraints(self, xs: jax.Array, cfs: jax.Array, hard: bool = False, 
                          rng_key: jrand.PRNGKey = None, **kwargs):
        return cfs
    
    def compute_reg_loss(self, xs, cfs, hard: bool = False):
        return 0.
    
    def from_dict(self, params: dict):
        self.name = params["name"]
        if not 'transformer' in params.keys():
            self.transformer = None
        else:
            self.transformer.from_dict(params["transformer"])
        return self
    
    def to_dict(self) -> dict:
        return {"name": self.name, "transformer": self.transformer.to_dict()}

# %% ../../nbs/data_utils/transform.ipynb 7
class MinMaxTransformation(_DefaultTransformation):
    def __init__(self):
        super().__init__("minmax", MinMaxScaler())

    def apply_constraints(self, xs, cfs, **kwargs):
        return jnp.clip(cfs, 0., 1.)

# %% ../../nbs/data_utils/transform.ipynb 9
class _OneHotTransformation(_DefaultTransformation):
    def __init__(self, name: str = None):
        super().__init__(name, OneHotEncoder())

    @property
    def num_categories(self) -> int:
        return len(self.transformer.categories_)
    
    def hard_constraints(self, operand: tuple[jax.Array, jrand.PRNGKey, dict]): 
        x, rng_key, kwargs = operand
        return jax.nn.one_hot(jnp.argmax(x, axis=-1), self.num_categories)
    
    def soft_constraints(self, operand: tuple[jax.Array, jrand.PRNGKey, dict]):
        raise NotImplementedError

    def apply_constraints(self, xs, cfs, hard: bool = False, rng_key=None, **kwargs):
        return jax.lax.cond(
            hard,
            true_fun=self.hard_constraints,
            false_fun=self.soft_constraints,
            operand=(cfs, rng_key, kwargs),
        )
    
    def compute_reg_loss(self, xs, cfs, hard: bool = False):
        reg_loss_per_xs = (cfs.sum(axis=-1, keepdims=True) - 1.0) ** 2
        return reg_loss_per_xs.mean()

# %% ../../nbs/data_utils/transform.ipynb 10
class SoftmaxTransformation(_OneHotTransformation):
    def __init__(self): 
        super().__init__("ohe")

    def soft_constraints(self, operand: tuple[jax.Array, jrand.PRNGKey, dict]):
        x, rng_key, kwargs = operand
        return jax.nn.softmax(x, axis=-1)
    
class GumbelSoftmaxTransformation(_OneHotTransformation):
    """Apply Gumbel softmax tricks for categorical transformation."""

    def __init__(self, tau: float = .1):
        super().__init__("gumbel")
        self.tau = tau
    
    def soft_constraints(self, operand: tuple[jax.Array, jrand.PRNGKey, dict]):
        x, rng_key, _ = operand
        if rng_key is None: # No randomness
            rng_key = jrand.PRNGKey(get_config().global_seed)
        return gumbel_softmax(rng_key, x, self.tau)
    
    def apply_constraints(self, xs, cfs, hard: bool = False, rng_key=None, **kwargs):
        """Apply constraints to the counterfactuals. If `rng_key` is None, no randomness is used."""
        return super().apply_constraints(xs, cfs, hard, rng_key, **kwargs)
    
    def to_dict(self) -> dict:
        return super().to_dict() | {"tau": self.tau}
    
def OneHotTransformation():
    warnings.warn("OneHotTransformation is deprecated since v0.2.5. "
                  "Use `SoftmaxTransformation`.", DeprecationWarning)
    return SoftmaxTransformation()

# %% ../../nbs/data_utils/transform.ipynb 12
class OrdinalTransformation(_DefaultTransformation):
    def __init__(self):
        super().__init__("ordinal", OrdinalPreprocessor())

    @property
    def num_categories(self) -> int:
        return len(self.transformer.categories_)
    
class IdentityTransformation(_DefaultTransformation):
    def __init__(self):
        super().__init__("identity", None)

    def fit(self, xs, y=None):
        return self
    
    def transform(self, xs):
        return xs
    
    def fit_transform(self, xs, y=None):
        return xs

    def apply_constraints(self, xs, cfs, **kwargs):
        return cfs
    
    def to_dict(self):
        return {'name': 'identity'}
    
    def from_dict(self, params: dict):
        self.name = params["name"]
        return self

# %% ../../nbs/data_utils/transform.ipynb 14
FEATURE_TRANSFORMATIONS = {
    'ohe': SoftmaxTransformation,
    'softmax': SoftmaxTransformation,
    'gumbel': GumbelSoftmaxTransformation,
    'minmax': MinMaxTransformation,
    'ordinal': OrdinalTransformation,
    'identity': IdentityTransformation,
}



"""Define utility funtions for `relax`."""

# AUTOGENERATED! DO NOT EDIT! File to edit: ../nbs/00_utils.ipynb.

# %% ../nbs/00_utils.ipynb 3
from __future__ import annotations
from .import_essentials import *
import nbdev
from fastcore.basics import AttrDict
from nbdev.showdoc import BasicMarkdownRenderer
from inspect import isclass
from fastcore.test import *
from jax.core import InconclusiveDimensionOperation

# %% auto 0
__all__ = ['validate_configs', 'save_pytree', 'load_pytree', 'auto_reshaping', 'grad_update', 'gumbel_softmax', 'load_json',
           'get_config', 'set_config']

# %% ../nbs/00_utils.ipynb 5
def validate_configs(
    configs: dict | BaseParser,  # A configuration of the model/dataset.
    config_cls: BaseParser,  # The desired configuration class.
) -> BaseParser:
    """return a valid configuration object."""

    assert isclass(config_cls), f"`config_cls` should be a class."
    assert issubclass(config_cls, BaseParser), \
        f"{config_cls} should be a subclass of `BaseParser`."
    
    if isinstance(configs, dict):
        configs = config_cls(**configs)
    if not isinstance(configs, config_cls):
        raise TypeError(
            f"configs should be either a `dict` or an instance of {config_cls.__name__}.")
    return configs

# %% ../nbs/00_utils.ipynb 15
def _is_array(x):
    return isinstance(x, np.ndarray) or isinstance(x, jnp.ndarray) or isinstance(x, list)

def save_pytree(pytree, saved_dir):
    """Save a pytree to a directory."""
    with open(os.path.join(saved_dir, "data.npy"), "wb") as f:
        for x in jax.tree_util.tree_leaves(pytree):
            np.save(f, x)

    tree_struct = jax.tree_util.tree_map(lambda t: _is_array(t), pytree)
    with open(os.path.join(saved_dir, "treedef.json"), "w") as f:
        json.dump(tree_struct, f)

# %% ../nbs/00_utils.ipynb 20
def load_pytree(saved_dir):
    """Load a pytree from a saved directory."""
    with open(os.path.join(saved_dir, "treedef.json"), "r") as f:
        tree_struct = json.load(f)

    leaves, treedef = jax.tree_util.tree_flatten(tree_struct)
    with open(os.path.join(saved_dir, "data.npy"), "rb") as f:
        flat_state = [
            np.load(f, allow_pickle=True) if is_arr else np.load(f, allow_pickle=True).item()
            for is_arr in leaves
        ]
    return jax.tree_util.tree_unflatten(treedef, flat_state)

# %% ../nbs/00_utils.ipynb 24
def _reshape_x(x: Array):
    x_size = x.shape
    if len(x_size) > 1 and x_size[0] != 1:
        raise ValueError(
            f"""Invalid Input Shape: Require `x.shape` = (1, k) or (k, ),
but got `x.shape` = {x.shape}. This method expects a single input instance."""
        )
    if len(x_size) == 1:
        x = x.reshape(1, -1)
    return x, x_size

# %% ../nbs/00_utils.ipynb 25
def auto_reshaping(
    reshape_argname: str, # The name of the argument to be reshaped.
    reshape_output: bool = True, # Whether to reshape the output. Useful to set `False` when returning multiple cfs.
):
    """
    Decorator to automatically reshape function's input into (1, k), 
    and out to input's shape.
    """
    def decorator(func):
        def wrapper(*args, **kwargs):
            kwargs = inspect.getcallargs(func, *args, **kwargs)
            if reshape_argname in kwargs:
                reshaped_x, x_shape = _reshape_x(kwargs[reshape_argname])
                kwargs[reshape_argname] = reshaped_x
            else:
                raise ValueError(
                    f"Invalid argument name: `{reshape_argname}` is not a valid argument name.")
            # Call the function.
            cf = func(**kwargs)
            if not isinstance(cf, Array): 
                raise ValueError(
                    f"Invalid return type: must be a `jax.Array`, but got `{type(cf).__name__}`.")
            if reshape_output:
                try: 
                    cf = cf.reshape(x_shape)
                except (InconclusiveDimensionOperation, TypeError) as e:
                    raise ValueError(
                        f"Invalid return shape: Require `cf.shape` = {cf.shape} "
                        f"is not compatible with `x.shape` = {x_shape}.")
            return cf

        return wrapper
    return decorator

# %% ../nbs/00_utils.ipynb 30
def grad_update(
    grads, # A pytree of gradients.
    params, # A pytree of parameters.
    opt_state: optax.OptState,
    opt: optax.GradientTransformation,
): # Return (upt_params, upt_opt_state)
    updates, opt_state = opt.update(grads, opt_state, params)
    upt_params = optax.apply_updates(params, updates)
    return upt_params, opt_state

# %% ../nbs/00_utils.ipynb 32
def gumbel_softmax(
    key: jrand.PRNGKey, # Random key
    logits: Array, # Logits for each class. Shape (batch_size, num_classes)
    tau: float, # Temperature for the Gumbel softmax
    axis: int | tuple[int, ...] = -1, # The axis or axes along which the gumbel softmax should be computed
):
    """The Gumbel softmax function."""

    gumbel_noise = jrand.gumbel(key, shape=logits.shape)
    y = logits + gumbel_noise
    return jax.nn.softmax(y / tau, axis=axis)

# %% ../nbs/00_utils.ipynb 34
def load_json(f_name: str) -> Dict[str, Any]:  # file name
    with open(f_name) as f:
        return json.load(f)


# %% ../nbs/00_utils.ipynb 36
@dataclass
class Config:
    rng_reserve_size: int
    global_seed: int

    @classmethod
    def default(cls) -> Config:
        return cls(rng_reserve_size=1, global_seed=42)

main_config = Config.default()

# %% ../nbs/00_utils.ipynb 37
def get_config() -> Config: 
    return main_config

# %% ../nbs/00_utils.ipynb 38
def set_config(
    *,
    rng_reserve_size: int = None, # The number of random number generators to reserve.
    global_seed: int = None, # The global seed for random number generators.
    **kwargs
) -> None:
    """Sets the global configurations."""

    def set_val(
        arg_name: str, # The name of the argument.
        arg_value: int, # The value of the argument.
        arg_min: int # The minimum value of the argument.
    ) -> None:
        """Checks the validity of the argument and sets the value."""
        
        if arg_value is None or not hasattr(main_config, arg_name):
            return
        
        if not isinstance(arg_value, int):
            raise TypeError(f"`{arg_name}` must be an integer, but got {type(arg_value).__name__}.")
        if arg_value < arg_min:
            raise ValueError(f"`{arg_name}` must be non-negative, but got {arg_value}.")
        setattr(main_config, arg_name, arg_value)

    set_val('rng_reserve_size', rng_reserve_size, 1)
    set_val('global_seed', global_seed, 0)



# AUTOGENERATED! DO NOT EDIT! File to edit: ../../nbs/methods/07_vaecf.ipynb.

# %% ../../nbs/methods/07_vaecf.ipynb 3
from __future__ import annotations
from ..import_essentials import *
from .base import ParametricCFModule
from ..ml_model import MLP, MLPBlock
from ..data_module import DataModule
from ..utils import auto_reshaping, validate_configs, get_config
from keras.random import SeedGenerator

# %% auto 0
__all__ = ['sample_latent', 'VAE', 'VAECFConfig', 'VAECF']

# %% ../../nbs/methods/07_vaecf.ipynb 4
@jax.jit
def hindge_embedding_loss(
    inputs: Array, targets: Array, margin: float = 1.0
):
    """Hinge embedding loss."""
    assert targets.shape == (1,)
    loss = jnp.where(
        targets == 1,
        inputs,
        jax.nn.relu(margin - inputs)
    )
    return loss

# %% ../../nbs/methods/07_vaecf.ipynb 7
def sample_latent(rng_key, mean, logvar):
    eps = jax.random.normal(rng_key, mean.shape)
    return mean + eps * jnp.sqrt(logvar)

# %% ../../nbs/methods/07_vaecf.ipynb 8
class VAE(keras.Model):
    def __init__(
        self, 
        layers: list[int],
        # pred_fn: Callable,
        mc_samples: int = 50,
        # compute_regularization_fn=None, 
        **kwargs
    ):
        super().__init__(**kwargs)
        self.n_layers = layers
        # self.pred_fn = pred_fn
        self.mc_samples = mc_samples
        self.rng_keys = SeedGenerator(get_config().global_seed)
        # if compute_regularization_fn is None:
        #     self.compute_regularization_fn = lambda *args, **kwargs: 0.
        # elif callable(compute_regularization_fn):
        #     self.compute_regularization_fn = compute_regularization_fn
        # else:
        #     raise ValueError("`compute_regularization_fn` must be callable or None, ",
        #                      f"but got {type(compute_regularization_fn)} instead.")
    
    def set_pred_fn(self, pred_fn):
        self.pred_fn = pred_fn

    def set_compute_regularization_fn(self, compute_regularization_fn):
        self.compute_regularization_fn = compute_regularization_fn

    def _compile(self, x):
        pred_out = self.pred_fn(x)
        if pred_out.shape[-1] != 2: 
            raise ValueError("Only binary classification is supported.")
        
        mu = self.mu_enc(x)
        var = 0.5 + self.var_enc(x)
        z = sample_latent(self.rng_keys.next(), mu, var)
        z = jnp.concatenate([z, pred_out.argmax(-1, keepdims=True)], axis=-1)
        mu_x = self.mu_dec(z)
    
    def build(self, input_shape):
        encoder = keras.Sequential([
            MLPBlock(size, use_batch_norm=True, dropout_rate=0.) for size in self.n_layers[:-1]
        ])
        decoder = keras.Sequential([
            MLPBlock(size, use_batch_norm=True, dropout_rate=0.) for size in self.n_layers[::-1][1:]
        ])

        self.mu_enc = keras.Sequential([encoder, keras.layers.Dense(self.n_layers[-1])])
        self.var_enc = keras.Sequential([encoder, keras.layers.Dense(self.n_layers[-1], activation='sigmoid')])
        self.mu_dec = keras.Sequential([
            decoder, keras.layers.Dense(input_shape[-1]), 
        ])
        self._compile(jnp.zeros(input_shape))

    def encode(self, x, training=None):
        mean = self.mu_enc(x, training=training)
        var = 0.5 + self.var_enc(x, training=training)
        return mean, var
    
    def decode(self, z, training=None):
        return self.mu_dec(z, training=training)
        
    def sample(
        self, 
        rng_key: jrand.PRNGKey, 
        inputs: Array, 
        mc_samples: int, 
        training=None
    ):
        @jit
        def step(rng_key, em, ev, c):
            # rng_key, _ = jrand.split(rng_key)
            z = sample_latent(rng_key, em, ev)
            z = jnp.concatenate([z, c], axis=-1)
            mu_x = self.decode(z)
            return mu_x

        keys = jrand.split(rng_key, mc_samples)
        x, c = inputs[:, :-1], inputs[:, -1:]
        em, ev = self.encode(x, training=training)
        step_fn = partial(step, em=em, ev=ev, c=c)
        mu_x = jax.vmap(step_fn)(keys) # [mc_samples, n, d]
        return em, ev, mu_x
    
    def compute_vae_loss(
        self,
        inputs: Array,
        em, ev, cfs
    ):
        def cf_loss(cf: Array, x: Array, y: Array):
            assert cf.shape == x.shape, f"cf.shape ({cf.shape}) != x.shape ({x.shape}))"
            # proximity loss
            recon_err = jnp.sum(jnp.abs(cf - x), axis=1).mean()
            # Sum to 1 over the categorical indexes of a feature
            cat_error = self.compute_regularization_fn(x, cf)
            # validity loss
            pred_prob = self.pred_fn(cf)
            # This is same as the following:
            # tempt_1, tempt_0 = pred_prob[y == 1], pred_prob[y == 0]
            # validity_loss = hindge_embedding_loss(tempt_1 - (1. - tempt_1), -1, 0.165) + \
            #     hindge_embedding_loss(1. - 2 * tempt_0, -1, 0.165)
            target = jnp.array([-1])
            hindge_loss_1 = hindge_embedding_loss(
                jax.nn.sigmoid(pred_prob[:, 1]) - jax.nn.sigmoid(pred_prob[:, 0]), target, 0.165)
            hindge_loss_0 = hindge_embedding_loss(
                jax.nn.sigmoid(pred_prob[:, 0]) - jax.nn.sigmoid(pred_prob[:, 1]), target, 0.165)
            tempt_1 = jnp.where(y == 1, hindge_loss_1, 0).sum() / y.sum()
            tempt_0 = jnp.where(y == 0, hindge_loss_0, 0).sum() / (y.shape[0] - y.sum())
            validity_loss = tempt_1 + tempt_0
            return recon_err + cat_error, - validity_loss
        
        xs, ys = inputs[:, :-1], inputs[:, -1]
        kl = 0.5 * jnp.mean(em**2 + ev - jnp.log(ev) - 1, axis=1)
        cf_loss_fn = partial(cf_loss, x=xs, y=ys)
        cf_losses, validity_losses = jax.vmap(cf_loss_fn)(cfs)
        return (cf_losses.mean() + kl).mean() + validity_losses.mean()
    
    def call(self, inputs, training=None):
        rng_key = self.rng_keys.next()
        ys = 1. - self.pred_fn(inputs).argmax(axis=1, keepdims=True)
        inputs = jnp.concatenate([inputs, ys], axis=-1)
        em, ev, cfs = self.sample(rng_key, inputs, self.mc_samples, training=training)
        loss = self.compute_vae_loss(inputs, em, ev, cfs)
        self.add_loss(loss)
        return cfs   


# %% ../../nbs/methods/07_vaecf.ipynb 9
class VAECFConfig(BaseParser):
    """Configurator of `VAECFModule`."""
    layers: List[int] = Field(
        [20, 16, 14, 12, 5],
        description="Sequence of Encoder/Decoder layer sizes."
    )
    dropout_rate: float = Field(
        0.1, description="Dropout rate."
    )
    opt_name: str = Field(
        "adam", description="Optimizer name."  
    )
    lr: float = Field(
        1e-3, description="Learning rate."
    )
    mc_samples: int = Field(
        50, description="Number of samples for mu."
    )
    validity_reg: float = Field(
        42.0, description="Regularization for validity."
    )


# %% ../../nbs/methods/07_vaecf.ipynb 10
class VAECF(ParametricCFModule):
    def __init__(self, config=None, vae=None, name: str = 'VAECF'):
        if config is None:
            config = VAECFConfig()
        config = validate_configs(config, VAECFConfig)
        self.vae = vae
        super().__init__(config, name=name)

    def _init_model(
        self, 
        config: VAECFConfig, 
        model: keras.Model, 
        # pred_fn: Callable,
        # compute_regularization_fn: Callable
    ):
        if model is None:
            model = VAE(
                config.layers,
                # pred_fn=pred_fn,
                mc_samples=config.mc_samples,
                # compute_regularization_fn=compute_regularization_fn
            )
            model.compile(
                optimizer=keras.optimizers.get({
                    'class_name': config.opt_name, 
                    'config': {'learning_rate': config.lr}
                }),
            )
        return model
    
    def train(
        self, 
        data: DataModule, 
        pred_fn: Callable, 
        batch_size: int = 128,
        epochs: int = 10,
        **fit_kwargs
    ):
        if not isinstance(data, DataModule):
            raise ValueError(f"Expected `data` to be `DataModule`, "
                             f"got type=`{type(data).__name__}` instead.")
        train_xs, train_ys = data['train']
        self.vae = self._init_model(self.config, self.vae)
        self.vae.set_pred_fn(pred_fn)
        self.vae.set_compute_regularization_fn(data.compute_reg_loss)
        self.vae.fit(
            train_xs, train_ys, 
            batch_size=batch_size, 
            epochs=epochs,
            **fit_kwargs
        )
        self._is_trained = True
        return self
    
    @auto_reshaping('x')
    def generate_cf(
        self,
        x: Array,
        pred_fn: Callable = None,
        y_target: Array = None,
        rng_key: jrand.PRNGKey = None,
        **kwargs
    ) -> Array:
        # TODO: Currently assumes binary classification.
        if y_target is None:
            y_target = 1 - pred_fn(x).argmax(axis=1, keepdims=True)
        else:
            y_target = y_target.reshape(1, -1).argmax(axis=1, keepdims=True)
        if rng_key is None:
            raise ValueError("`rng_key` must be provided, but got `None`.")
                
        @jit
        def sample_step(rng_key, y_target):
            inputs = jnp.concatenate([x, y_target], axis=-1)
            _, _, cfs = self.vae.sample(rng_key, inputs, 1, training=False)
            return cfs
        
        return sample_step(rng_key, y_target)
        



# AUTOGENERATED! DO NOT EDIT! File to edit: ../../nbs/methods/08_clue.ipynb.

# %% ../../nbs/methods/08_clue.ipynb 3
from __future__ import annotations
from ..import_essentials import *
from .base import CFModule, ParametricCFModule
from ..base import BaseConfig
from ..utils import auto_reshaping, validate_configs, get_config, grad_update
from ..ml_model import MLP, MLPBlock
from ..data_module import DataModule
from jax.scipy.stats.norm import logpdf as gaussian_logpdf
from keras.random import SeedGenerator


# %% auto 0
__all__ = ['Encoder', 'Decoder', 'kl_divergence', 'VAEGaussCat', 'CLUEConfig', 'get_reconstruction_loss_fn', 'CLUE']

# %% ../../nbs/methods/08_clue.ipynb 5
class Encoder(keras.layers.Layer):
    def __init__(self, sizes: List[int], dropout: float = 0.1):
        super().__init__()
        assert sizes[-1] % 2 == 0, f"sizes[-1] must be even, but got {sizes[-1]}"
        self.encoder = keras.Sequential([
            MLPBlock(size, dropout_rate=dropout) for size in sizes
        ])
    
    def call(self, x: Array, training: bool):
        params = self.encoder(x, training=training)
        d = params.shape[-1] // 2
        mu, sigma = params[:, :d], params[:, d:]
        sigma = jax.nn.softplus(sigma)
        sigma = jnp.clip(sigma, 1e-3)
        return mu, sigma

class Decoder(keras.layers.Layer):
    def __init__(self, sizes: List[int], output_size: int, dropout: float = 0.1):
        super().__init__()
        self.decoder = MLP(
            sizes, output_size=output_size, 
            dropout_rate=dropout, last_activation='sigmoid'
        )
    
    def __call__(self, z: Array, training: bool):
        mu_dec = self.decoder(z, training=training)
        return mu_dec

# %% ../../nbs/methods/08_clue.ipynb 7
@jit
def kl_divergence(p: Array, q: Array, eps: float = 2 ** -17) -> Array:
    loss_pointwise = p * (jnp.log(p + eps) - jnp.log(q + eps))
    return loss_pointwise

# %% ../../nbs/methods/08_clue.ipynb 8
class VAEGaussCat(keras.Model):
    def __init__(
        self, 
        enc_sizes: List[int] = [20, 16, 14, 12],
        dec_sizes: List[int] = [12, 14, 16, 20],
        dropout_rate: float = 0.1,    
    ):
        super().__init__()
        self.enc_sizes = enc_sizes
        self.dec_sizes = dec_sizes
        self.dropout_rate = dropout_rate
        self.seed_generator = SeedGenerator(get_config().global_seed)
        # default reconstruction loss to l2 loss
        self.reconstruction_loss = lambda x, y: optax.l2_loss(x, y).mean(-1)

    def set_reconstruction_loss(self, fn):
        self.reconstruction_loss = fn

    def build(self, input_shape):
        self.encoder = Encoder(self.enc_sizes, self.dropout_rate)
        self.decoder = Decoder(self.dec_sizes, input_shape[-1], self.dropout_rate)
    
    def encode(self, x, training=True):
        mu_z, var_z = self.encoder(x, training=training)
        return mu_z, var_z
    
    def sample_latent(self, rng_key, mean, var):
        key, _ = jax.random.split(rng_key)
        std = jnp.exp(0.5 * var)
        eps = jrand.normal(key, var.shape)
        return mean + eps * std
    
    def decode(self, z, training=True):
        reconstruct_x = self.decoder(z, training=training)
        return reconstruct_x        
    
    def sample_step(self, rng_key, mean, var, training=True):
        z = self.sample_latent(rng_key, mean, var)
        mu_x = self.decode(z, training=training)
        return mu_x
    
    def sample(self, x, mc_samples, training=True): # Shape: (mc_samples, batch_size, input_size)
        mean, var = self.encode(x, training=training)
        rng_key = self.seed_generator.next()
        keys = jax.random.split(rng_key, mc_samples)
        
        partial_sample_step = partial(
            self.sample_step, mean=mean, var=var, training=training
        )
        reconstruct_x = jax.vmap(partial_sample_step)(keys)
        return (mean, var, reconstruct_x)
        
    def compute_vae_loss(self, inputs, mu_z, logvar_z, reconstruct_x):
        kl_loss = -0.5 * (1 + logvar_z - jnp.power(mu_z, 2) - jnp.exp(logvar_z)).sum(-1)
        
        rec = self.reconstruction_loss(inputs, reconstruct_x.reshape(inputs.shape)).sum(-1)
        batchwise_loss = (rec + kl_loss) / inputs.shape[0]
        return batchwise_loss.mean()

    def call(self, inputs, training=True):
        mu_z, logvar_z, reconstruct_x = self.sample(inputs, mc_samples=1, training=training)
        loss = self.compute_vae_loss(inputs, mu_z, logvar_z, reconstruct_x)
        self.add_loss(loss)
        return reconstruct_x

# %% ../../nbs/methods/08_clue.ipynb 10
@ft.partial(jit, static_argnums=(3, 4, 6, 9, 12, 13))
def _clue_generate(
    x: Array,
    rng_key: jrand.PRNGKey,
    y_target: Array,
    pred_fn: Callable,
    max_steps: int,
    step_size: float,
    vae_module: VAEGaussCat,
    uncertainty_weight: float,
    aleatoric_weight: float,
    prior_weight: float,
    distance_weight: float,
    validity_weight: float,
    validity_fn: Callable,
    apply_fn: Callable
) -> Array:
    
    @jit
    def sample_latent_from_x(x: Array, rng_key: jrand.PRNGKey):
        key_1, key_2 = jrand.split(rng_key)
        mean, var = vae_module.encode(x, training=False)
        z = vae_module.sample_latent(key_2, mean, var)
        return z
    
    @ft.partial(jit, static_argnums=(1))
    def generate_from_z(z: Array, hard: bool):
        cf = vae_module.decode(z, training=False)
        cf = apply_fn(x, cf, hard=hard)
        return cf

    @jit
    def uncertainty_from_z(z: Array):
        cfs = generate_from_z(z, hard=False)
        pred_cfs = pred_fn(cfs)
        prob = pred_cfs[:, 1]
        total_uncertainty = -(prob * jnp.log(prob + 1e-10)).sum(-1)
        return total_uncertainty, cfs, pred_cfs
    
    @jit
    def compute_loss(z: Array):
        uncertainty, cfs, pred_cfs = uncertainty_from_z(z)
        loglik = gaussian_logpdf(z).sum(-1)
        dist = jnp.abs(cfs - x).mean()
        validity = validity_fn(y_target, pred_cfs).mean()
        loss = (
            (uncertainty_weight + aleatoric_weight) * uncertainty 
            + prior_weight * loglik
            + distance_weight * dist
            + validity_weight * validity
        )
        return loss.mean()
    
    @loop_tqdm(max_steps)
    def step(i, z_opt_state):
        z, opt_state = z_opt_state
        z_grad = jax.grad(compute_loss)(z)
        z, opt_state = grad_update(z_grad, z, opt_state, opt)
        return z, opt_state
    
    key_1, _ = jax.random.split(rng_key)
    z = sample_latent_from_x(x, key_1)
    opt = optax.adam(step_size)
    opt_state = opt.init(z)

    # Write a loop to optimize z using lax.fori_loop
    z, opt_state = lax.fori_loop(0, max_steps, step, (z, opt_state))
    cf = generate_from_z(z, hard=True)
    return cf

# %% ../../nbs/methods/08_clue.ipynb 12
class CLUEConfig(BaseConfig):
    enc_sizes: List[int] = Field(
        [20, 16, 14, 12], description="Sequence of Encoder layer sizes."
    )
    dec_sizes: List[int] = Field(
        [12, 14, 16, 20], description="Sequence of Decoder layer sizes."
    )
    dropout_rate: float = Field(0.1, description="Dropout rate")
    encoded_size: int = Field(5, description="Encoded size")
    lr: float = Field(0.001, description="Learning rate")
    max_steps: int = Field(500, description="Max steps")
    step_size: float = Field(0.01, description="Step size")
    vae_n_epochs: int = Field(10, description="Number of epochs for VAE")
    vae_batch_size: int = Field(128, description="Batch size for VAE")
    seed: int = Field(0, description="Seed for random number generator")

# %% ../../nbs/methods/08_clue.ipynb 13
def get_reconstruction_loss_fn(dm: DataModule):
    def reconstruction_loss(xs, cfs):
        losses = []
        for feat, (start, end) in dm.features.features_and_indices:
            if feat.is_categorical:
                losses.append(
                    optax.softmax_cross_entropy(cfs[:, start:end], xs[:, start:end]).reshape(-1, 1)
                )
            else:
                losses.append(optax.l2_loss(cfs[:, start:end], xs[:, start:end]))
        return jnp.concatenate(losses, axis=-1)
    
    return reconstruction_loss

# %% ../../nbs/methods/08_clue.ipynb 15
class CLUE(ParametricCFModule):

    def __init__(self, config: Dict | CLUEConfig = None, vae=None, name: str = 'CLUE'):
        if config is None:
            config = CLUEConfig()
        config = validate_configs(config, CLUEConfig)
        self.vae = vae
        super().__init__(config, name=name)

    def _init_model(self, config: CLUEConfig, model: VAEGaussCat):
        if model is None:
            model = VAEGaussCat(
                enc_sizes=config.enc_sizes, dec_sizes=config.dec_sizes, 
                dropout_rate=config.dropout_rate
            )
            model.compile(optimizer=keras.optimizers.Adam(config.lr), loss=None)
        return model
    
    def train(
        self, 
        data: DataModule, # data module
        pred_fn: Callable = None,
        batch_size: int = 128,
        epochs: int = 10,
        **fit_kwargs
    ):
        if not isinstance(data, DataModule):
            raise ValueError(f"Expected `data` to be `DataModule`, "
                             f"got type=`{type(data).__name__}` instead.")
        X_train, y_train = data['train'] 
        self.vae = self._init_model(self.config, self.vae)
        self.vae.fit(
            X_train, X_train, 
            batch_size=batch_size, 
            epochs=epochs,
            **fit_kwargs
        )
        self._is_trained = True
        return self

    @auto_reshaping('x')
    def generate_cf(
        self,
        x: Array,
        pred_fn: Callable,
        y_target: Array = None,
        rng_key: jrand.PRNGKey = None,
        **kwargs
    ) -> Array:
        # TODO: Currently assumes binary classification.
        if y_target is None:
            y_target = 1 - pred_fn(x)
        else:
            y_target = jnp.array(y_target, copy=True)
        if rng_key is None:
            raise ValueError("`rng_key` must be provided, but got `None`.")
        return _clue_generate(
            x, 
            rng_key=rng_key, 
            y_target=y_target,
            pred_fn=pred_fn,
            max_steps=self.config.max_steps,
            step_size=self.config.step_size,
            vae_module=self.vae,
            uncertainty_weight=.0,
            aleatoric_weight=0.0,
            prior_weight=0.0,
            distance_weight=.1,
            validity_weight=1.0,
            validity_fn=keras.losses.get({'class_name': 'KLDivergence', 'config': {'reduction': None}}),
            apply_fn=self.apply_constraints,
        )



__version__ = "0.2.8"

from .data_module import DataModule, DataModuleConfig, load_data
from .data_utils import (
    Feature, FeaturesList
)
from .ml_model import MLModule, MLModuleConfig, load_ml_module
from .explain import generate_cf_explanations
from .evaluate import evaluate_cfs, benchmark_cfs



# AUTOGENERATED! DO NOT EDIT! File to edit: ../../nbs/methods/06_cchvae.ipynb.

# %% ../../nbs/methods/06_cchvae.ipynb 3
from __future__ import annotations
from ..import_essentials import *
from .base import CFModule, ParametricCFModule
from ..base import BaseConfig
from ..utils import auto_reshaping, grad_update, validate_configs, get_config
from ..data_utils import Feature, FeaturesList
from ..ml_model import MLP, MLPBlock
from ..data_module import DataModule
from keras.random import SeedGenerator

# %% auto 0
__all__ = ['CHVAE', 'CCHVAEConfig', 'CCHVAE']

# %% ../../nbs/methods/06_cchvae.ipynb 5
@keras.saving.register_keras_serializable()
class CHVAE(keras.Model):
    def __init__(
        self, 
        layers: list, 
        dropout_rate: float = 0., 
        **kwargs
    ):
        super().__init__()
        self.n_layers = layers
        self.dropout_rate = dropout_rate
        self.seed_generator = SeedGenerator(get_config().global_seed)

    def set_apply_constraints_fn(self, fn=None):
        if fn is None:
            fn = lambda x: x
        self.apply_constraints = fn

    def build(self, input_shape):
        encoder = keras.Sequential([
            MLPBlock(size, dropout_rate=self.dropout_rate) for size in self.n_layers[:-1]
        ])
        decoder = keras.Sequential([
            MLPBlock(size, dropout_rate=self.dropout_rate) for size in self.n_layers[::-1][1:]
        ])
        # Encoder
        self.mu_enc = keras.Sequential([encoder, keras.layers.Dense(self.n_layers[-1])])
        self.log_var_enc = keras.Sequential([encoder, keras.layers.Dense(self.n_layers[-1])])
        # Decoder
        self.mu_dec = keras.Sequential([
            decoder, keras.layers.Dense(input_shape[-1]), 
            keras.layers.BatchNormalization(),
        ])
        self.log_var_dec = keras.Sequential([
            decoder, keras.layers.Dense(input_shape[-1]), 
            keras.layers.BatchNormalization(),
        ])

    def encode(self, x, training=None):
        return self.mu_enc(x, training=training), self.log_var_enc(x, training=training)
    
    def decode(self, z, training=None):
        return self.mu_dec(z, training=training), self.log_var_dec(z, training=training)
    
    def reparameterize(self, mu, log_var):
        std = keras.ops.exp(0.5 * log_var)
        eps = keras.random.normal(std.shape, seed=self.seed_generator)
        return mu + eps * std
    
    def forward(self, x, training=None):
        mu_z, log_var_z = self.encode(x, training=training)
        z = self.reparameterize(mu_z, log_var_z)
        mu_x, log_var_x = self.decode(z, training=training)
        return mu_z, log_var_z, z, mu_x, log_var_x

    def call(self, inputs, training=None):
        x = inputs
        mu_z, log_var_z, z, mu_x, log_var_x = self.forward(x, training=training)
        # compute loss
        loss = self.compute_kl_loss(mu_z, log_var_z)
        self.add_loss(loss)
        reconstructed_x = mu_x
        # reconstructed_x = self.apply_constraints(inputs, mu_x, hard=not training)
        return reconstructed_x
    
    def regenerate(self, z):
        mu, log_var = self.decode(z)
        return mu
    
    def compute_kl_loss(self, mu, logvar):
        kl_loss = -0.5 * keras.ops.sum(1 + logvar - mu**2 - keras.ops.exp(logvar))
        return kl_loss

# %% ../../nbs/methods/06_cchvae.ipynb 6
def _hyper_sphere_coordindates(
    rng_key: jrand.PRNGKey, # Random number generator key
    x: Array, # Input instance with only continuous features. Shape: (1, n_features)
    n_samples: int,
    high: float, # Upper bound
    low: float, # Lower bound
    p_norm: int = 2 # Norm
):
    key_1, key_2 = jrand.split(rng_key)
    delta = jrand.normal(key_1, shape=(n_samples, x.shape[-1]))
    dist = jrand.normal(key_2, shape=(n_samples,)) * (high - low) + low
    norm_p = jnp.linalg.norm(delta, ord=p_norm, axis=1)
    d_norm = jnp.divide(dist, norm_p).reshape(-1, 1)  # rescale/normalize factor
    delta = jnp.multiply(delta, d_norm)
    candidates = x + delta
    return candidates
     

@ft.partial(jit, static_argnums=(3, 4, 5, 7))
def _cchvae(
    x: Array,
    rng_key: jrand.PRNGKey,
    y_target: Array,
    pred_fn: Callable[[Array], Array],
    max_steps: int,
    n_search_samples: int,
    step_size: float,
    chvae: CHVAE,
):
    """Counterfactual generation using CCHVAE."""
        
    @loop_tqdm(max_steps)
    def body_fn(i, state):
        count, candidate_cf, rng = state
        rng_key, subkey_1, subkey_2 = jrand.split(rng, num=3)
        low, high = step_size * count, step_size * (count + 1)
        # STEP 1 -- SAMPLE POINTS on hyper sphere around instance
        latent_neighbors = _hyper_sphere_coordindates(
            subkey_1, z_rep, n_search_samples, high=high, low=low, p_norm=1
        )
        x_ce = chvae.regenerate(latent_neighbors)
        x_ce = chvae.apply_constraints(x, x_ce.reshape(1, -1), hard=True)
        
        # STEP 2 -- COMPUTE l1 norms
        distances = jnp.abs(x_ce - x).sum(axis=1)

        # STEP 3 -- SELECT POINT with MINIMUM l1 norm
        y_candidates = pred_fn(x_ce).argmax(axis=1)
        indices = jnp.where(y_candidates == y_target, 1, 0).astype(bool)
        distances = jnp.where(indices, distances, jnp.inf)

        best_candidate_cf = x_ce[jnp.argmin(distances)].reshape(1, -1)
        
        candidate_cf = lax.cond(
            distances.min() < jnp.abs(x - candidate_cf).sum(axis=1).min(),
            lambda _: best_candidate_cf,
            lambda _: candidate_cf,
            None
        )

        count += 1
        return count, candidate_cf, rng_key
    
    y_target = y_target.reshape(1, -1).argmax(axis=1)
    z, _ = chvae.encode(x)
    # z_rep = jnp.repeat(z.reshape(1, -1), n_search_samples, axis=0)
    z_rep = z.reshape(1, -1)
    rng_key, _ = jrand.split(rng_key)
    # candidate_cf = jnp.array(x, copy=True)
    candidate_cf = jnp.ones_like(x) * jnp.inf
    state = (0, candidate_cf, rng_key) # (count, candidate_cf, rng_key)
    # count, candidate_cf, rng_key = jax.lax.while_loop(cond_fn, body_fn, state)
    count, candidate_cf, rng_key = lax.fori_loop(0, max_steps, body_fn, state)
    # if `inf` is found, return the original input
    candidate_cf = jnp.where(jnp.isinf(candidate_cf), x, candidate_cf)
    return candidate_cf


# %% ../../nbs/methods/06_cchvae.ipynb 7
class CCHVAEConfig(BaseConfig):
    vae_layers: List[int] = Field(
        [20, 16, 14, 12], description="List of hidden layer sizes for VAE."
    )
    opt_name: str = Field("adam", description="Optimizer name of VAE.")
    vae_lr: float = Field(0.001, description="Learning rate of VAE.")
    max_steps: int = Field(100, description="Max steps")
    n_search_samples: int = Field(100, description="Number of generated candidate counterfactuals.")
    step_size: float = Field(0.1, description="Step size")    

# %% ../../nbs/methods/06_cchvae.ipynb 8
class CCHVAE(ParametricCFModule):
    
    def __init__(self, config: Dict | CCHVAEConfig = None, chvae=None, name: str = 'cchvae'):
        if config is None:
            config = CCHVAEConfig()
        config = validate_configs(config, CCHVAEConfig)
        name = "CCHVAE" if name is None else name
        self.vae = chvae
        super().__init__(config, name=name)

    def _init_model(self, config: CCHVAEConfig, model: keras.Model):
        if model is None:
            model = CHVAE(self.config.vae_layers)
            model.compile(
                optimizer=keras.optimizers.get({
                    'class_name': config.opt_name, 
                    'config': {'learning_rate': config.vae_lr}
                }),
                loss=keras.losses.MeanSquaredError()
            )
        return model

    def train(
        self, 
        data: DataModule, 
        pred_fn: Callable = None,
        batch_size: int = 128,
        epochs: int = 10,
        **fit_kwargs
    ):
        if not isinstance(data, DataModule):
            raise ValueError(f"Expected `data` to be `DataModule`, "
                             f"got type=`{type(data).__name__}` instead.")
        X_train, y_train = data['train'] 
        self.vae = self._init_model(self.config, self.vae)
        self.vae.set_apply_constraints_fn(data.apply_constraints)
        self.vae.fit(
            X_train, X_train, 
            batch_size=batch_size, 
            epochs=epochs,
            **fit_kwargs
        )
        self._is_trained = True
        return self
    
    @auto_reshaping('x')
    def generate_cf(
        self,
        x: Array,
        pred_fn: Callable = None,
        y_target: Array = None,
        rng_key: jrand.PRNGKey = None,
        **kwargs
    ) -> Array:
        # TODO: Currently assumes binary classification.
        if y_target is None:
            y_target = 1 - pred_fn(x)
        else:
            y_target = jnp.array(y_target, copy=True)
        if rng_key is None:
            raise ValueError("`rng_key` must be provided, but got `None`.")
        
        return _cchvae(
            x,
            rng_key=rng_key,
            y_target=y_target,
            pred_fn=pred_fn,
            max_steps=self.config.max_steps,
            n_search_samples=self.config.n_search_samples,
            step_size=self.config.step_size,
            chvae=self.vae,
        )



# AUTOGENERATED! DO NOT EDIT! File to edit: ../nbs/03_explain.ipynb.

# %% ../nbs/03_explain.ipynb 2
from __future__ import annotations
from .import_essentials import *
from .data_module import DataModule, load_data
from .base import *
from .methods import *
from .strategy import *
from .ml_model import *
from .utils import get_config, save_pytree, load_pytree
import einops
from sklearn.datasets import make_classification
from copy import deepcopy

# %% auto 0
__all__ = ['Explanation', 'fake_explanation', 'prepare_pred_fn', 'prepare_cf_module', 'prepare_rng_keys',
           'generate_cf_explanations']

# %% ../nbs/03_explain.ipynb 4
class Explanation(DataModule):
    """Generated CF Explanations class. It inherits a `DataModule`."""

    def __init__(
        self,
        cfs: Array,  # Generated cf explanation of `xs` in `data`
        pred_fn: Callable[[Array], Array],  # Predict function
        data_module: DataModule = None,  # Data module
        xs: Array = None,  # Input data
        ys: Array = None,  # Target data
        total_time: float = None,  # Total runtime
        cf_name: str = "CFModule",  # CF method's name
        data=None, # Deprecated argument
    ):
        if data is not None:
            warnings.warn(
                "Argument `data` is deprecated. Use `data_module` instead.",
                DeprecationWarning,
            )
            data_module = data

        if (xs is None or ys is None) and data_module is None:
            raise ValueError(
                "Either `xs` and `ys` or `data_module` must be provided."
            )
        
        if data_module is None:
            data_module = DataModule.from_numpy(xs, ys, transformation='identity')
        # assign attributes
        # self.recourses = data_module.features.with_transformed_data(cfs)
        self._cfs = cfs
        self.pred_fn = pred_fn
        self.total_time = total_time
        self.cf_name = cf_name
        
        super().__init__(
            features=data_module.features, 
            label=data_module.label,
            config=data_module.config,
            data=data_module.data,
        )

    def __repr__(self):
        return f"Explanation(data_name={self.data_name}, cf_name={self.cf_name}, " \
               f"total_time={self.total_time}, xs={self.xs}, ys={self.ys}, cfs={self.cfs})"

    def __getitem__(self, name: Literal['train', 'val', 'test']) -> Dict[str, Array]:
        if name == 'train':
            indices = self.train_indices
        elif name in ['val', 'test']:
            indices = self.test_indices
        else:
            raise ValueError(f"Unknown data name: {name}. Should be one of ['train', 'val', 'test']")

        if isinstance(indices, list):
            indices = jnp.array(indices)
        
        return {
            'xs': self.xs[indices],
            'ys': self.ys[indices],
            'cfs': self.cfs[indices],
        }    

    @property
    def cfs(self) -> Array:
        """Return the counterfactuals in the shape of (n, c, k)"""
        if self._cfs.ndim == 2:
            return einops.rearrange(self._cfs, "n d -> n () d")
        return self._cfs
    
    @property
    def data_name(self):
        return self.name
    
    @property
    def feature_indices(self):
        return self.features.feature_indices
    
    @property
    def features_and_indices(self):
        return self.features.features_and_indices
    
    def copy(self):
        """Return a deep copy of the explanation. 
        
        Warning: this method will not create a deepcopy of `pred_fn`.
        """

        dm = DataModule(self.features, self.label, self.config, self.data)
        return Explanation(
            cfs=self.cfs.copy(),
            pred_fn=self.pred_fn,
            data_module=dm,
            total_time=self.total_time,
            cf_name=self.cf_name,
        )
        
    def save(self, path: str):
        """Save the explanation to a directory."""
        # create directories
        dm_path = Path(path) / 'data'
        exp_path = Path(path) / 'explanations'
        exp_path.mkdir(parents=True, exist_ok=True)
        # save data module and explanations
        super().save(dm_path)        
        save_pytree({
            'cfs': self.cfs,
            'total_time': self.total_time,
            'cf_name': self.cf_name,
        }, exp_path)
    
    @classmethod
    def load_from_path(cls, path: str, *, ml_module_path: str = None):
        dm_path = Path(path) / 'data'
        exp_path = Path(path) / 'explanations'
        dm = DataModule.load_from_path(dm_path)
        explanations = load_pytree(exp_path)
        if ml_module_path is not None:
            pred_fn = MLModule.load_from_path(ml_module_path).pred_fn
        else:
            warnings.warn("`ml_module_path` is not provided. Setting `pred_fn=None`.")
            pred_fn = None
        return cls(
            pred_fn=pred_fn,
            data_module=dm,
            **explanations
        )

    __ALL__ = [
        'copy', 'save', 'load_from_path'
    ]

# %% ../nbs/03_explain.ipynb 5
def fake_explanation(n_cfs: int=1):
    dm = load_data('dummy')
    ml_model = load_ml_module('dummy')
    if n_cfs < 1: 
        raise ValueError(f'n_cfs must be greater than 0, but got n_cfs={n_cfs}.')
    elif n_cfs == 1:
        cfs = dm.xs
    else:
        # Allow for multiple counterfactuals
        cfs = einops.repeat(dm.xs, "n k -> n c k", c=n_cfs)

    return Explanation(
        data_module=dm, cfs=cfs, pred_fn=ml_model.pred_fn, total_time=0.0, cf_name='dummy_method'
    )

# %% ../nbs/03_explain.ipynb 10
def prepare_pred_fn(
    cf_module: CFModule,
    data: DataModule,
    pred_fn: Callable[[Array, ...], Array], # Predictive function. 
    pred_fn_args: Dict = None,
) -> Callable[[Array], Array]: # Return predictive function with signature `(x: Array) -> Array`.
    """Prepare the predictive function for the CF module. 
    We will train the model if `pred_fn` is not provided and `cf_module` does not have `pred_fn`.
    If `pred_fn` is found in `cf_module`, we will use it irrespective of `pred_fn` argument.
    If `pred_fn` is provided, we will use it.
    """
    # Train the model if `pred_fn` is not provided.
    if not hasattr(cf_module, 'pred_fn') and pred_fn is None:
        model = MLModule().train(data)
        return model.pred_fn
    # If `pred_fn` is detected in cf_module, 
    # use it irrespective of `pred_fn` argument.
    elif hasattr(cf_module, 'pred_fn'):
        return cf_module.pred_fn
    # If `pred_fn` is provided, use it.
    else:
        if pred_fn_args is not None:
            pred_fn = ft.partial(pred_fn, **pred_fn_args)
        return pred_fn

def prepare_cf_module(
    cf_module: CFModule,
    data_module: DataModule,
    pred_fn: Callable[[Array], Array] = None,
    train_config: Dict[str, Any] = None, 
):
    """Prepare the CF module. 
    It will hook up the data module, 
    and its apply functions via the `init_apply_fns` method
    (e.g., `apply_constraints_fn` and `compute_reg_loss_fn`).
    Next, it will train the model if `cf_module` is a `ParametricCFModule`.
    Finally, it will call `before_generate_cf` method.
    """
    cf_module.set_data_module(data_module)
    cf_module.set_apply_constraints_fn(data_module.apply_constraints)
    cf_module.set_compute_reg_loss_fn(data_module.compute_reg_loss)
    train_config = train_config or {}
    if isinstance(cf_module, ParametricCFModule):
        if not cf_module.is_trained:
            cf_module.train(data_module, pred_fn=pred_fn, **train_config)
    cf_module.before_generate_cf()
    return cf_module

def prepare_rng_keys(
    rng_key: jrand.PRNGKey,
    n_instances: int,
):
    """Prepare random number generator keys."""
    if rng_key is None:
        rng_key = jrand.PRNGKey(get_config().global_seed)
    rng_keys = jrand.split(rng_key, n_instances)
    return rng_keys


# %% ../nbs/03_explain.ipynb 11
def generate_cf_explanations(
    cf_module: CFModule, # CF Explanation Module
    data: DataModule, # Data Module
    pred_fn: Callable[[Array, ...], Array] = None, # Predictive function
    strategy: str | BaseStrategy = None, # Parallelism Strategy for generating CFs. Default to `vmap`.
    train_config: Dict[str, Any] = None, 
    pred_fn_args: dict = None, # auxiliary arguments for `pred_fn` 
    rng_key: jrand.PRNGKey = None, # Random number generator key
) -> Explanation: # Return counterfactual explanations.
    """Generate CF explanations."""

    # Prepare `pred_fn`, `cf_module`, and `strategy`.
    pred_fn = prepare_pred_fn(cf_module, data, pred_fn, pred_fn_args)
    cf_module = prepare_cf_module(cf_module, data, pred_fn, train_config)
    if strategy is None:
        strategy = StrategyFactory.get_default_strategy()
    strategy = StrategyFactory.get_strategy(strategy)
    # n_instances
    n_instances = data.xs.shape[0]
    # Prepare random number generator keys.
    rng_keys = prepare_rng_keys(rng_key, n_instances)
    y_targets = 1 - pred_fn(data.xs)
    
    # Generate CF explanations.
    start_time = time.time()
    cfs = strategy(cf_module.generate_cf, data.xs, pred_fn, y_targets, rng_keys)
    # cfs = jax.vmap(cf_module.generate_cf, in_axes=(0, None, 0, 0))(data.xs, pred_fn, y_targets, rng_keys)
    total_time = time.time() - start_time

    # Return CF explanations.
    return Explanation(
        cf_name=cf_module.name,
        data=data,
        cfs=cfs,
        total_time=total_time,
        pred_fn=pred_fn,
    )



# AUTOGENERATED! DO NOT EDIT! File to edit: ../../nbs/methods/03_proto.ipynb.

# %% ../../nbs/methods/03_proto.ipynb 3
from __future__ import annotations
from ..import_essentials import *
from ..base import TrainableMixedin, BaseConfig
from .base import ParametricCFModule
from ..utils import validate_configs, auto_reshaping, grad_update
from ..ml_model import AutoEncoder
from ..data_module import DataModule

# %% auto 0
__all__ = ['ProtoCFConfig', 'ProtoCF']

# %% ../../nbs/methods/03_proto.ipynb 5
@ft.partial(jit, static_argnums=(2, 3, 9, 10, 12))
def _proto_cf(
    x: Array, 
    y_target: Array,
    pred_fn: Callable[[Array], Array],
    n_steps: int,
    lr: float,
    c: float, # hyperparameter for validity loss
    beta: float, # cost = beta *l1_norm + l2_norm
    gamma: float, # hyperparameter for loss_ae
    theta: float, # hyperparameter for loss_proto
    ae: keras.Model,
    validity_fn: Callable,
    sampled_data: Array,
    apply_constraints_fn: Callable,
) -> Array:
    
    @jit
    def encode(x):
        return ae.encoder(x)
    
    @jit
    def cost_fn(cf, x):
        # For some reasons, calling jnp.linalg.norm(cf - x) 
        # directly will lead to significant performance drop.
        return beta * jnp.abs(cf - x).mean() + optax.l2_loss(cf, x).mean()
    
    @ft.partial(jit, static_argnums=(3))
    def loss_fn(
        cf: Array,
        x: Array,
        y_target: Array,
        pred_fn: Callable[[Array], Array],
    ):
        y_cf = pred_fn(cf)
        loss_val = c * validity_fn(y_target, y_cf)
        # loss_cost = beta * jnp.linalg.norm(cf - x, ord=1) + jnp.linalg.norm(cf - x, ord=2)
        loss_cost = cost_fn(cf, x)
        loss_ae = gamma * jnp.square(ae(cf) - cf).mean()
        loss_proto = theta * jnp.square(
            jnp.linalg.norm(encode(cf) - encode(sampled_data).sum(axis=0) / n_sampled_data, ord=2)
        )
        return (loss_val + loss_cost + loss_ae + loss_proto).mean()
    
    @loop_tqdm(n_steps)
    def gen_cf_step(
        i, cf_opt_state: Tuple[Array, optax.OptState] 
    ) -> Tuple[Array, optax.OptState]:
        cf, opt_state = cf_opt_state
        cf_grads = jax.grad(loss_fn)(cf, x, y_target, pred_fn)
        cf, opt_state = grad_update(cf_grads, cf, opt_state, opt)
        cf = apply_constraints_fn(x, cf, hard=False)
        return cf, opt_state
    
    # Calculate the number of samples
    # If the sampled data is all zeros, which means that this is not a valid sample.
    # This is used to calculate the mean of encode(sampled_data)
    n_sampled_data = jnp.where((sampled_data == 0).all(axis=1), 0, 1).sum()
    cf = jnp.array(x, copy=True)
    opt = optax.adam(lr)
    opt_state = opt.init(cf)
    cf, opt_state = lax.fori_loop(0, n_steps, gen_cf_step, (cf, opt_state))
    cf = apply_constraints_fn(x, cf, hard=True)
    return cf

# %% ../../nbs/methods/03_proto.ipynb 6
class ProtoCFConfig(BaseConfig):
    """Configurator of `ProtoCF`."""
    
    n_steps: int = 100
    lr: float = 0.01
    c: float = Field(1, description="The weight for validity loss.")
    beta: float = Field(0.1, description="The weight for l1_norm in the cost function, where cost = beta * l1_norm + l2_norm.")
    gamma: float = Field(0.1, description="The weight for Autoencoder loss.")
    theta: float = Field(0.1, description="The weight for prototype loss.")
    n_samples: int = Field(128, description="Number of samples for prototype.")
    validity_fn: str = 'KLDivergence'
    # AE configs
    enc_sizes: List[int] = Field([64, 32, 16], description="List of hidden layers of Encoder.")
    dec_sizes: List[int] = Field([16, 32, 64], description="List of hidden layers of Decoder.")
    opt_name: str = Field("adam", description="Optimizer name of AutoEncoder.")
    ae_lr: float = Field(1e-3, description="Learning rate of AutoEncoder.")
    ae_loss: str = Field("mse", description="Loss function name of AutoEncoder.")


# %% ../../nbs/methods/03_proto.ipynb 7
class ProtoCF(ParametricCFModule):

    def __init__(
        self,
        config: dict | ProtoCFConfig = None,
        ae: keras.Model = None,
        name: str = None,
    ):
        if config is None:
            config = ProtoCFConfig()
        config = validate_configs(config, ProtoCFConfig)
        self.ae = ae
        name = "ProtoCF" if name is None else name
        super().__init__(config, name=name)

    def _init_model(self, config: ProtoCFConfig, model: keras.Model, output_size: int):
        if model is None:
            model = AutoEncoder(
                enc_sizes=config.enc_sizes,
                dec_sizes=config.dec_sizes,
                output_size=output_size,
            )
            model.compile(
                optimizer=keras.optimizers.get({
                    'class_name': config.opt_name, 
                    'config': {'learning_rate': config.ae_lr}
                }),
                loss=config.ae_loss,
            )
        return model
    
    def train(
        self, 
        data: DataModule, 
        pred_fn: Callable = None,
        batch_size: int = 128,
        epochs: int = 10,
        **fit_kwargs
    ):
        if not isinstance(data, DataModule):
            raise ValueError(f"Expected `data` to be `DataModule`, got type=`{type(data).__name__}` instead.")
        X_train, y_train = data['train'] 
        self.ae = self._init_model(self.config, self.ae, X_train.shape[1])
        self.ae.fit(
            X_train, X_train, 
            batch_size=batch_size, 
            epochs=epochs,
            **fit_kwargs
        )
        self._is_trained = True
        # self.sampled_data = data.sample(self.config.n_samples)
        sampled_xs, sampled_ys = data.sample(self.config.n_samples)
        self.sampled_data = tuple(map(jax.device_put, (sampled_xs, sampled_ys)))
        return self
    
    @auto_reshaping('x')
    def generate_cf(
        self,
        x: Array,  # `x` shape: (k,), where `k` is the number of features
        pred_fn: Callable[[Array], Array],
        y_target: Array = None,
        **kwargs,
    ) -> Array:
        # TODO: Select based on the closest prototype.
        if y_target is None:
            y_target = 1 - pred_fn(x)
        else:
            y_target = jnp.array(y_target, copy=True).reshape(1, -1)

        sampled_data = jnp.where(
            y_target.argmax(axis=1) == self.sampled_data[1],
            self.sampled_data[0],
            jnp.zeros_like(self.sampled_data[0]),
        )

        return _proto_cf(
            x=x,
            y_target=y_target,
            pred_fn=pred_fn,
            n_steps=self.config.n_steps,
            lr=self.config.lr,
            c=self.config.c,
            beta=self.config.beta,
            gamma=self.config.gamma,
            theta=self.config.theta,
            ae=self.ae,
            sampled_data=sampled_data,
            validity_fn=keras.losses.get({'class_name': self.config.validity_fn, 'config': {'reduction': None}}),
            apply_constraints_fn=self.apply_constraints,
        )




# AUTOGENERATED! DO NOT EDIT! File to edit: ../nbs/04_evaluate.ipynb.

# %% ../nbs/04_evaluate.ipynb 3
from __future__ import annotations

from .import_essentials import *
from .base import *
from .explain import *
from keras.metrics import sparse_categorical_accuracy
import einops

# %% auto 0
__all__ = ['BaseEvalMetrics', 'PredictiveAccuracy', 'compute_single_validity', 'compute_validity', 'Validity',
           'compute_single_proximity', 'compute_proximity', 'Proximity', 'compute_single_sparsity', 'compute_sparsity',
           'Sparsity', 'ManifoldDist', 'Runtime', 'evaluate_cfs', 'benchmark_cfs']

# %% ../nbs/04_evaluate.ipynb 6
class BaseEvalMetrics:
    """Base evaluation metrics class."""

    def __init__(self, name: str = None):
        if name is None: 
            name = type(self).__name__
        self.name = name

    def __str__(self) -> str:
        has_name = hasattr(self, 'name')
        if not has_name:
            raise ValidationError(
                "EvalMetrics must have a name. Add the following as the first line in your "
                f"__init__ method:\n\nsuper({self.__name__}, self).__init__()")
        return self.name

    def __call__(self, explanation: Explanation) -> Any:
        raise NotImplementedError


# %% ../nbs/04_evaluate.ipynb 7
class PredictiveAccuracy(BaseEvalMetrics):
    """Compute the accuracy of the predict function."""
    
    def __init__(self, name: str = "accuracy"):
        super().__init__(name=name)

    def __call__(self, explanation: Explanation) -> float:
        xs, ys = explanation.xs, explanation.ys
        pred_fn = explanation.pred_fn
        pred_ys = pred_fn(xs)
        accuracy = sparse_categorical_accuracy(ys, pred_ys)
        return accuracy.mean()

# %% ../nbs/04_evaluate.ipynb 9
def compute_single_validity(
    xs: Array, # (n, d)
    cfs: Array, # (n, d)
    pred_fn: Callable[[Array], Array],
):
    y_xs = pred_fn(xs).argmax(axis=-1)
    y_cfs = pred_fn(cfs).argmax(axis=-1)
    validity = 1 - jnp.equal(y_xs, y_cfs).mean()
    return validity

def compute_validity(
    xs: Array, # (n, d)
    cfs: Array, # (n, d) or (n, b, d)
    pred_fn: Callable[[Array], Array],
) -> float:
    cfs = einops.rearrange(cfs, 'n ... d -> n (...) d')
    valdity_batch = jax.vmap(compute_single_validity, in_axes=(None, 1, None))(xs, cfs, pred_fn)
    return valdity_batch.mean()

# %% ../nbs/04_evaluate.ipynb 11
class Validity(BaseEvalMetrics):
    """Compute fraction of input instances on which CF explanation methods output valid CF examples.
    Support binary case only.
    """
    
    def __init__(self, name: str = "validity"):
        super().__init__(name=name)

    def __call__(self, explanation: Explanation) -> float:
        xs, cfs, pred_fn = explanation.xs, explanation.cfs, explanation.pred_fn
        return compute_validity(xs, cfs, pred_fn)

# %% ../nbs/04_evaluate.ipynb 13
def compute_single_proximity(xs: Array, cfs: Array):
    prox = jnp.linalg.norm(xs - cfs, ord=1, axis=1).mean()
    return prox

def compute_proximity(xs: Array, cfs: Array) -> float:
    cfs = einops.rearrange(cfs, 'n ... d -> n (...) d')
    prox_batch = jax.vmap(compute_single_proximity, in_axes=(None, 1))(xs, cfs)
    return prox_batch.mean()

# %% ../nbs/04_evaluate.ipynb 15
class Proximity(BaseEvalMetrics):
    """Compute L1 norm distance between input datasets and CF examples divided by the number of features."""
    def __init__(self, name: str = "proximity"):
        super().__init__(name=name)
    
    def __call__(self, explanation: Explanation) -> float:
        xs, cfs = explanation.xs, explanation.cfs
        return compute_proximity(xs, cfs)

# %% ../nbs/04_evaluate.ipynb 17
def compute_single_sparsity(xs: Array, cfs: Array, feature_indices: List[Tuple[int, int]]):
    def _feat_sparsity(xs, cfs, feat_indices):
        start, end = feat_indices
        xs = xs[:, start: end]
        cfs = cfs[:, start: end]
        return jnp.linalg.norm(xs - cfs, ord=0, axis=1).mean()
    
    return jnp.stack([_feat_sparsity(xs, cfs, feat_indices) for feat_indices in feature_indices]).mean()

def compute_sparsity(xs: Array, cfs: Array, feature_indices: List[Tuple[int, int]]) -> float:
    cfs = einops.rearrange(cfs, 'n ... d -> n (...) d')
    sparsity_batch = jax.vmap(compute_single_sparsity, in_axes=(None, 1, None))(xs, cfs, feature_indices)
    return sparsity_batch.mean()

# %% ../nbs/04_evaluate.ipynb 18
class Sparsity(BaseEvalMetrics):
    """Compute the number of feature changes between input datasets and CF examples."""

    def __init__(self, name: str = "sparsity"):
        super().__init__(name=name)
    
    def __call__(self, explanation: Explanation) -> float:
        xs, cfs, feature_indices = explanation.xs, explanation.cfs, explanation.feature_indices
        return compute_sparsity(xs, cfs, feature_indices)

# %% ../nbs/04_evaluate.ipynb 20
@partial(jit, static_argnums=(2))
def pairwise_distances(
    x: Array, # [n, k]
    y: Array, # [m, k]
    metric: str = "euclidean" # Supports "euclidean" and "cosine"
) -> Array: # [n, m]
    def euclidean_distances(x: Array, y: Array) -> float:
        XX = jnp.dot(x, x)
        YY = jnp.dot(y, y)
        XY = jnp.dot(x, y)
        dist = jnp.clip(XX - 2 * XY + YY, a_min=0.)
        return jnp.sqrt(dist)
        # return jnp.linalg.norm(x - y, ord=2)
    
    def cosine_distances(x: Array, y: Array) -> float:
        return 1.0 - jnp.dot(x, y) / (jnp.linalg.norm(x) * jnp.linalg.norm(y) + 1e-8)
    
    if metric == "euclidean":
        dists_fn = vmap(vmap(euclidean_distances, in_axes=(None, 0)), in_axes=(0, None))
    elif metric == "cosine":
        dists_fn = vmap(vmap(cosine_distances, in_axes=(None, 0)), in_axes=(0, None))
    else:
        raise ValueError(f"metric='{metric}' not supported")
    
    return dists_fn(x, y)

# %% ../nbs/04_evaluate.ipynb 21
@ft.partial(jax.jit, static_argnames=["k", "recall_target"])
def l2_ann(
    qy, # Query vectors
    db, # Database
    k=10, # Number of nearest neighbors to return
    recall_target=0.95 # Recall target for the approximation.
) -> Tuple[Array, Array]: # Return (distance, neighbor_indices) tuples
    dists = pairwise_distances(qy, db)
    return jax.lax.approx_min_k(dists, k=k, recall_target=recall_target)

# %% ../nbs/04_evaluate.ipynb 22
class ManifoldDist(BaseEvalMetrics):
    """Compute the L1 distance to the n-nearest neighbor for all CF examples."""
    def __init__(self, n_neighbors: int = 1, name: str = "manifold_dist"):
        super().__init__(name=name)
        self.n_neighbors = n_neighbors
        
    def __call__(self, explanation: Explanation) -> float:
        xs, cfs = explanation.xs, explanation.cfs
        l2_ann_partial = ft.partial(l2_ann, k=self.n_neighbors)
        dists, _ = vmap(l2_ann_partial, in_axes=(1, None))(cfs, xs)
        return dists.mean()

# %% ../nbs/04_evaluate.ipynb 24
class Runtime(BaseEvalMetrics):
    """Compute the runtime of the CF explanation method."""
    def __init__(self, name: str = "runtime"):
        super().__init__(name=name)
    
    def __call__(self, explanation: Explanation) -> float:
        return explanation.total_time

# %% ../nbs/04_evaluate.ipynb 27
METRICS_CALLABLE = [
    PredictiveAccuracy('acc'),
    PredictiveAccuracy('accuracy'),
    Validity(),
    Proximity(),
    Runtime(),
    ManifoldDist(),
]

METRICS = { m.name: m for m in METRICS_CALLABLE }

DEFAULT_METRICS = ["acc", "validity", "proximity"]

# %% ../nbs/04_evaluate.ipynb 29
def _get_metric(metric: str | BaseEvalMetrics, cf_exp: Explanation):
    if isinstance(metric, str):
        if metric not in METRICS.keys():
            raise ValueError(f"'{metric}' is not supported. Must be one of {METRICS.keys()}")
        res = METRICS[metric](cf_exp)
    elif callable(metric):
        # f(cf_exp) not supported for now
        if not isinstance(metric, BaseEvalMetrics):
            raise ValueError(f"metric needs to be a subclass of `BaseEvalMetrics`.")
        res = metric(cf_exp)
    else:
        raise ValueError(f"{type(metric).__name__} is not supported as a metric.")
    
    # Get scalar value
    if isinstance(res, Array) and res.ravel().shape == (1,):
        res = res.item()
    return res


# %% ../nbs/04_evaluate.ipynb 31
def evaluate_cfs(
    cf_exp: Explanation, # CF Explanations
    metrics: Iterable[Union[str, BaseEvalMetrics]] = None, # A list of Metrics. Can be `str` or a subclass of `BaseEvalMetrics`
    return_dict: bool = True, # return a dictionary or not (default: True)
    return_df: bool = False # return a pandas Dataframe or not (default: False)
):
    cf_name = cf_exp.cf_name
    data_name = cf_exp.data_name
    result_dict = { (data_name, cf_name): dict() }

    if metrics is None:
        metrics = DEFAULT_METRICS

    for metric in metrics:
        metric_name = str(metric)
        result_dict[(data_name, cf_name)][metric_name] = _get_metric(metric, cf_exp)
    result_df = pd.DataFrame.from_dict(result_dict, orient="index")
    
    if return_dict and return_df:
        return (result_dict, result_df)
    elif return_dict or return_df:
        return result_df if return_df else result_dict


# %% ../nbs/04_evaluate.ipynb 33
def benchmark_cfs(
    cf_results_list: Iterable[Explanation],
    metrics: Optional[Iterable[str]] = None,
):
    dfs = [
        evaluate_cfs(
            cf_exp=cf_results, metrics=metrics, return_dict=False, return_df=True
        )
        for cf_results in cf_results_list
    ]
    return pd.concat(dfs)




"""`DataModule` for training parametric models, generating and benchmarking CF explanations."""

# AUTOGENERATED! DO NOT EDIT! File to edit: ../nbs/01_data.ipynb.

# %% ../nbs/01_data.ipynb 3
from __future__ import annotations
from .utils import load_json, validate_configs, get_config, save_pytree, load_pytree, get_config
from .base import *
from .data_utils import *
from .import_essentials import *
import jax
from jax import numpy as jnp, random as jrand, lax, Array
import pandas as pd
import numpy as np
from pathlib import Path
import json, os, shutil
from urllib.request import urlretrieve
from pydantic.fields import ModelField, Field
from typing import List, Dict, Union, Optional, Tuple, Callable, Any, Iterable
import warnings
from pandas.testing import assert_frame_equal

# %% auto 0
__all__ = ['BaseDataModule', 'DataModuleConfig', 'features2config', 'features2pandas', 'dataframe2features', 'dataframe2labels',
           'DataModule', 'download_data_module_files', 'load_data']

# %% ../nbs/01_data.ipynb 6
class BaseDataModule(BaseModule):
    """DataModule Interface"""

    def _prepare(self, *args, **kwargs):
        """Prepare data for training"""
        raise NotImplementedError
        
    def apply_constraints(self, x: Array, cf: Array, hard: bool = False, **kwargs) -> Array:
        raise NotImplementedError
    
    def compute_reg_loss(self, x: Array, cf: Array, hard: bool = False, **kwargs) -> float:
        raise NotImplementedError

# %% ../nbs/01_data.ipynb 7
class DataModuleInfoMixin:
    """This base class exposes some attributes of DataModule
    at the base level for easy access.
    """

    @property
    def data(self) -> pd.DataFrame:
        return self._data
    
    @property
    def xs(self) -> Array:
        return self._features.transformed_data
    
    @property
    def ys(self) -> Array:
        return self._label.transformed_data
    
    @property
    def features(self) -> FeaturesList:
        return self._features
    
    @property
    def label(self) -> FeaturesList:
        return self._label

    @property
    def dataset(self) -> Tuple[Array, Array]:
        return (self.xs, self.ys)
    
    @property
    def train_indices(self) -> List[int]:
        return self.config.train_indices
    
    @property
    def test_indices(self) -> List[int]:
        return self.config.test_indices


# %% ../nbs/01_data.ipynb 10
class DataModuleConfig(BaseConfig):
    """Configurator of `DataModule`."""

    data_dir: str = Field(None, description="The directory of dataset.")
    data_name: str = Field(None, description="The name of `DataModule`.")
    continous_cols: List[str] = Field([], description="Continuous features/columns in the data.")
    discret_cols: List[str] = Field([], description="Categorical features/columns in the data.")
    imutable_cols: List[str] = Field([], description="Immutable features/columns in the data.")
    continuous_transformation: Optional[str] = Field('minmax', description="Transformation for continuous features. `None` indicates unknown.")
    discret_transformation: Optional[str] = Field('ohe', description="Transformation for categorical features. `None` indicates unknown.")
    sample_frac: Optional[float] = Field(
        None, description="Sample fraction of the data. Default to use the entire data.", ge=0., le=1.0
    )
    train_indices: List[int] = Field([], description="Indices of training data.")
    test_indices: List[int] = Field([], description="Indices of testing data.")
    
    def shuffle(self, data: Array, test_size: float, seed: int = None):
        """Shuffle data with a seed."""
        if seed is None:
            seed = get_config().global_seed
        key = jrand.PRNGKey(seed)
        total_length = data.shape[0]
        train_length = int((1 - test_size) * total_length)
        if len(self.train_indices) == 0:
            self.train_indices = jrand.permutation(key, total_length)[:train_length].tolist()
        if len(self.test_indices) == 0:
            self.test_indices = jrand.permutation(key, total_length)[train_length:].tolist()

# %% ../nbs/01_data.ipynb 13
def features2config(
    features: FeaturesList, # FeaturesList to be converted
    name: str, # Name of the data used for `DataModuleConfig`
    return_dict: bool = False # Whether to return a dict or `DataModuleConfig`
) -> Union[DataModuleConfig, Dict]: # Return configs
    """Get `DataModuleConfig` from `FeaturesList`."""

    cont, cats, immu = [], [], []
    cont_transformation, cat_transformation = None, None
    for f in features:
        if f.is_categorical:
            cats.append(f.name)
        else:
            cont.append(f.name)
        if f.is_immutable:
            immu.append(f.name)
    
    configs_dict = {
        "data_dir": ".",
        "data_name": name,
        "continous_cols": cont,
        "discret_cols": cats,
        "imutable_cols": immu,
        "continuous_transformation": cont_transformation,
        "discret_transformation": cat_transformation,
    }
    if return_dict:
        return configs_dict
    return DataModuleConfig(**configs_dict)


# %% ../nbs/01_data.ipynb 15
def features2pandas(
    features: FeaturesList, # FeaturesList to be converted
    labels: FeaturesList # labels to be converted
) -> pd.DataFrame: # Return pandas dataframe
    """Convert `FeaturesList` to pandas dataframe."""
    
    feats_df = features.to_pandas()
    labels_df = labels.to_pandas()
    df = pd.concat([feats_df, labels_df], axis=1)
    return df

# %% ../nbs/01_data.ipynb 18
def to_feature(col: str, data: pd.DataFrame, config: DataModuleConfig, transformation: str):
    return Feature(
        name=col, data=data[col].to_numpy().reshape(-1, 1),
        transformation=transformation,
        is_immutable=col in config.imutable_cols
    )

# %% ../nbs/01_data.ipynb 19
def dataframe2features(
    data: pd.DataFrame,
    config: DataModuleConfig,
) -> FeaturesList:
    """Convert pandas dataframe of features to `FeaturesList`."""

    cont_features = [to_feature(col, data, config, config.continuous_transformation) for col in config.continous_cols]
    cat_features = [to_feature(col, data, config, config.discret_transformation) for col in config.discret_cols]
    features = cont_features + cat_features
    return FeaturesList(features)


def dataframe2labels(
    data: pd.DataFrame,
    config: DataModuleConfig,
) -> FeaturesList:
    """Convert pandas dataframe of labels to `FeaturesList`."""
    
    label_cols = set(data.columns) - set(config.continous_cols) - set(config.discret_cols)
    labels = [to_feature(col, data, config, 'identity') for col in label_cols]
    return FeaturesList(labels)

# %% ../nbs/01_data.ipynb 21
class DataModule(BaseDataModule, DataModuleInfoMixin):
    """DataModule for tabular data."""

    def __init__(
        self, 
        features: FeaturesList,
        label: FeaturesList,
        config: DataModuleConfig = None,
        data: pd.DataFrame = None,
        **kwargs
    ):
        self._prepare(features, label)
        if config is None:
            name = kwargs.pop('name', 'DataModule')
            config = features2config(features, name)
        config.shuffle(self.xs, test_size=0.25)
        self._data = features2pandas(features, label) if data is None else data
        super().__init__(config, name=config.data_name)

    def _prepare(self, features, label):
        if features is not None and label is not None:
            self._features = FeaturesList(features)
            self._label = FeaturesList(label)
        elif features is None:
            raise ValueError("Features cannot be None.")
        elif label is None:
            raise ValueError("Label cannot be None.")
            
    def save(
        self, 
        path: str # Path to the directory to save `DataModule`
    ):
        """Save `DataModule` to a directory."""
        path = Path(path)
        if not path.exists():
            path.mkdir(parents=True)
        self._features.save(path / 'features')
        self._label.save(path / 'label')
        if self._data is not None:
            self._data.to_csv(path / 'data.csv', index=False)
        with open(path / "config.json", "w") as f:
            json.dump(self.config.dict(), f)

    @classmethod
    def load_from_path(
        cls, 
        path: str,  # Path to the directory to load `DataModule`
        config: Dict|DataModuleConfig = None # Configs of `DataModule`. This argument is ignored.
    ) -> DataModule: # Initialized `DataModule` from path
        """Load `DataModule` from a directory."""
        if config is not None:
            warnings.warn("Passing `config` will have no effect.")
        
        path = Path(path)
        config = DataModuleConfig.load_from_json(path / 'config.json')
        # config = validate_configs(config, DataModuleConfig)
        features = FeaturesList.load_from_path(path / 'features')
        label = FeaturesList.load_from_path(path / 'label')
        data = pd.read_csv(path / 'data.csv')
        return cls(features=features, label=label, config=config, data=data)
    
    @classmethod
    def from_path(cls, path, config: DataModuleConfig = None):
        """Alias of `load_from_path`"""
        return cls.load_from_path(path, config)
    
    @classmethod
    def from_config(
        cls, 
        config: Dict|DataModuleConfig, # Configs of `DataModule`
        data: pd.DataFrame=None # Passed in pd.Dataframe
    ) -> DataModule: # Initialized `DataModule` from configs and data
        config = validate_configs(config, DataModuleConfig)
        if data is None:
            data = pd.read_csv(config.data_dir)
        if not isinstance(data, pd.DataFrame):
            raise ValueError("`data` should be a pandas DataFrame.")
        features = dataframe2features(data, config)
        label = dataframe2labels(data, config)
        return cls(features=features, label=label, config=config, data=data)
    
    @classmethod
    def from_numpy(
        cls,
        xs: np.ndarray, # Input data
        ys: np.ndarray, # Labels
        name: str = None, # Name of `DataModule`
        transformation='minmax'
    ) -> DataModule: # Initialized `DataModule` from numpy arrays
        """Create `DataModule` from numpy arrays. Note that the `xs` are treated as continuous features."""
        
        features = FeaturesList([Feature(f"feature_{i}", xs[:, i].reshape(-1, 1), transformation=transformation) for i in range(xs.shape[1])])
        labels = FeaturesList([Feature(f"label", ys.reshape(-1, 1), transformation='identity')])
        return cls(features=features, label=labels, name=name)
    
    @classmethod
    def from_features(
        cls, 
        features: FeaturesList, # Features of `DataModule`
        label: FeaturesList, # Labels of `DataModule`
        name: str = None # Name of `DataModule`
    ) -> DataModule: # Initialized `DataModule` from features and labels
        """Create `DataModule` from `FeaturesList`."""
        return cls(features=features, label=label, name=name)
        
    def _get_data(self, indices):
        if isinstance(indices, list):
            indices = jnp.array(indices)
        return (self.xs[indices], self.ys[indices])
        
    def __getitem__(self, name: str):
        if name == 'train':
            return self._get_data(self.config.train_indices)
        elif name in ['valid', 'test']:
            return self._get_data(self.config.test_indices)
        else:
            raise ValueError(f"Unknown data name: {name}. Should be one of ['train', 'valid', 'test']")
    
    def set_transformations(
        self, 
        feature_names_to_transformation: Dict[str, Union[str, Dict, BaseTransformation]], # Dict[feature_name, Transformation]
    ) -> DataModule:
        """Reset transformations for features."""

        self._features = self._features.set_transformations(feature_names_to_transformation)
        return self
    
    def sample(
        self, 
        size: float | int, # Size of the sample. If float, should be 0<=size<=1.
        stage: str = 'train', # Stage of data to sample from. Should be one of ['train', 'valid', 'test']
        key: jrand.PRNGKey = None # Random key. 
    ) -> Tuple[Array, Array]: # Sampled data
        """Sample data from `DataModule`."""

        key = jrand.PRNGKey(get_config().global_seed) if key is None else key
        xs, ys = self[stage]
        indices = jnp.arange(xs.shape[0])
        
        if isinstance(size, float) and 0 <= size <= 1:
            size = int(size * indices.shape[0])
        elif isinstance(size, int):
            size = min(size, indices.shape[0])
        else:
            raise ValueError(f"`size` should be a floating number 0<=size<=1, or an integer,"
                             f" but got size={size}.")
                
        indices = jrand.permutation(key, indices)[:size]
        return xs[indices], ys[indices]

    def transform(
        self, 
        data: pd.DataFrame | Dict[str, Array] # Data to be transformed
    ) -> Array: # Transformed data
        """Transform data to `jax.Array`."""
        # TODO: test this function
        if isinstance(data, pd.DataFrame):
            data_dict = {k: np.array(v).reshape(-1, 1) for k, v in data.iloc[:, :-1].to_dict(orient='list').items()}
            return self._features.transform(data_dict)
        elif isinstance(data, dict):
            data = jax.tree_util.tree_map(lambda x: np.array(x).reshape(-1, 1), data)
            return self._features.transform(data)
        else:
            raise ValueError("data should be a pandas DataFrame or `Dict[str, jax.Array]`.")
        
    def inverse_transform(
        self, 
        data: Array, # Data to be inverse transformed
        return_type: str = 'pandas' # Type of the returned data. Should be one of ['pandas', 'dict']
    ) -> pd.DataFrame: # Inverse transformed data
        """Inverse transform data to `pandas.DataFrame`."""
        # TODO: test this function
        inversed = self._features.inverse_transform(data)
        if return_type == 'pandas':
            return inversed
        elif return_type == 'dict':
            raise NotImplementedError
        else:
            raise ValueError(f"Unknown return type: {return_type}. Should be one of ['pandas', 'dict']")
        
    def apply_constraints(
        self, 
        xs: Array, # Input data
        cfs: Array, # Counterfactuals to be constrained
        hard: bool = False, # Whether to apply hard constraints or not
        rng_key: jrand.PRNGKey = None, # Random key
        **kwargs
    ) -> Array: # Constrained counterfactuals
        """Apply constraints to counterfactuals."""
        return self._features.apply_constraints(xs, cfs, hard=hard, rng_key=rng_key, **kwargs)
    
    def compute_reg_loss(
        self, 
        xs: Array, # Input data
        cfs: Array, # Counterfactuals to be constrained
        hard: bool = False # Whether to apply hard constraints or not
    ) -> float:
        """Compute regularization loss."""
        return self._features.compute_reg_loss(xs, cfs, hard)
    
    __ALL__ = [
        'load_from_path', 
        'from_config', 
        'from_features',
        'from_numpy',
        'save',
        'transform',
        'inverse_transform',
        'apply_constraints',
        'compute_reg_loss',
        'set_transformations',
        'sample'
    ]

# %% ../nbs/01_data.ipynb 22
def dm_equals(dm1: DataModule, dm2: DataModule):
    # data_equals = np.allclose(dm1.data.to_numpy(), dm2.data.to_numpy())
    assert_frame_equal(dm1.data, dm2.data)
    xs_equals = np.allclose(dm1.xs, dm2.xs)
    ys_equals = np.allclose(dm1.ys, dm2.ys)
    train_indices_equals = np.array_equal(dm1.train_indices, dm2.train_indices)
    test_indices_equals = np.array_equal(dm1.test_indices, dm2.test_indices)
    # print(f"data_equals: {data_equals}, xs_equals: {xs_equals}, ys_equals: {ys_equals}, train_indices_equals: {train_indices_equals}, test_indices_equals: {test_indices_equals}")
    return (
        xs_equals and ys_equals and 
        train_indices_equals and test_indices_equals
    )

# %% ../nbs/01_data.ipynb 30
DEFAULT_DATA = [
    'adult',
    'heloc',
    'oulad',
    'credit',
    'cancer',
    'student_performance',
    'titanic',
    'german',
    'spam',
    'ozone',
    'qsar',
    'bioresponse',
    'churn',
    'road',
    'dummy'
 ]

DEFAULT_DATA_CONFIGS = { 
    data: { 
        'data': f"{data}/data", 'model': f"{data}/model",
    } for data in DEFAULT_DATA
}

# %% ../nbs/01_data.ipynb 35
def _validate_dataname(data_name: str):
    if data_name not in DEFAULT_DATA:
        raise ValueError(f'`data_name` must be one of {DEFAULT_DATA}, '
            f'but got data_name={data_name}.')

# %% ../nbs/01_data.ipynb 36
def download_data_module_files(
    data_name: str, # The name of data
    data_parent_dir: Path, # The directory to save data.
    download_original_data: bool = False, # Download original data or not
):
    files = [
        "features/data.npy", "features/treedef.json",
        "label/data.npy", "label/treedef.json",
        "config.json"
    ]
    if download_original_data:
        files.append("data.csv")
    for f in files:
        url = f"https://huggingface.co/datasets/birkhoffg/ReLax-Assets/resolve/main/{data_name}/data/{f}"
        f_path = data_parent_dir / f'{data_name}/data' / f
        os.makedirs(f_path.parent, exist_ok=True)
        if not f_path.is_file(): urlretrieve(url, f_path)


def load_data(
    data_name: str, # The name of data
    return_config: bool = False, # Deprecated
    data_configs: dict = None, # Data configs to override default configuration
) -> DataModule | Tuple[DataModule, DataModuleConfig]: # Return `DataModule` or (`DataModule`, `DataModuleConfig`)
    """High-level util function for loading `data` and `data_config`."""
    
    _validate_dataname(data_name)

    # create new dir
    data_parent_dir = Path(os.getcwd()) / "relax-assets"
    if not data_parent_dir.exists():
        os.makedirs(data_parent_dir, exist_ok=True)
    # download files
    download_data_module_files(
        data_name, data_parent_dir, 
        download_original_data=True
    )

    if return_config:
        warnings.warn("`return_config` is deprecated since v0.2. "
                      "Please access `config` from `DataModule`.", DeprecationWarning)

    # read and override config
    # comment them for now since we cannot garantee the override configs are valid
    # conf_path = data_parent_dir / f'{data_name}/data/config.json'
    # config = load_json(conf_path)
    # if not (data_configs is None):
    #     config.update(data_configs)
    # config = DataModuleConfig(**config)

    data_dir = data_parent_dir / f'{data_name}/data'
    data_module = DataModule.load_from_path(data_dir, config=data_configs)

    return data_module




# AUTOGENERATED! DO NOT EDIT! File to edit: ../../nbs/methods/01_vanilla.ipynb.

# %% ../../nbs/methods/01_vanilla.ipynb 3
from __future__ import annotations
from ..import_essentials import *
from .base import CFModule
from ..base import BaseConfig
from ..utils import auto_reshaping, grad_update, validate_configs

# %% auto 0
__all__ = ['VanillaCFConfig', 'VanillaCF']

# %% ../../nbs/methods/01_vanilla.ipynb 5
@ft.partial(jit, static_argnums=(2, 3, 6, 7, 8))
def _vanilla_cf(
    x: jnp.DeviceArray,  # `x` shape: (k,), where `k` is the number of features
    y_target: Array, # `y_target` shape: (1,)
    pred_fn: Callable[[Array], Array],  # y = pred_fn(x)
    n_steps: int,
    lr: float,  # learning rate for each `cf` optimization step
    lambda_: float,  #  loss = validity_loss + lambda_params * cost
    validity_fn: Callable,
    cost_fn: Callable,
    apply_constraints_fn: Callable
) -> jnp.DeviceArray:  # return `cf` shape: (k,)
    @jit
    def loss_fn_1(y_true: Array, y_pred: Array):
        return validity_fn(y_true, y_pred).mean()

    @jit
    def loss_fn_2(x: Array, cf: Array):
        return cost_fn(cf, x).mean()

    @partial(jit, static_argnums=(2,))
    def loss_fn(
        cf: Array,  # `cf` shape: (k, 1)
        x: Array,  # `x` shape: (k, 1)
        pred_fn: Callable[[Array], Array],
    ):
        cf_y_pred = pred_fn(cf)
        return loss_fn_1(y_target, cf_y_pred) + lambda_ * loss_fn_2(x, cf)

    @loop_tqdm(n_steps)
    def gen_cf_step(
        i, cf_opt_state: Tuple[Array, optax.OptState] #x: Array, cf: Array, opt_state: optax.OptState
    ) -> Tuple[jnp.DeviceArray, optax.OptState]:
        cf, opt_state = cf_opt_state
        cf_grads = jax.grad(loss_fn)(cf, x, pred_fn)
        cf, opt_state = grad_update(cf_grads, cf, opt_state, opt)
        cf = apply_constraints_fn(x, cf, hard=False)
        return cf, opt_state

    cf = jnp.array(x, copy=True)
    opt = optax.rmsprop(lr)
    opt_state = opt.init(cf)
    cf, opt_state = lax.fori_loop(0, n_steps, gen_cf_step, (cf, opt_state))

    cf = apply_constraints_fn(x, cf, hard=True)
    return cf

# %% ../../nbs/methods/01_vanilla.ipynb 6
class VanillaCFConfig(BaseConfig):
    n_steps: int = 100
    lr: float = 0.1
    lambda_: float = 0.1
    validity_fn: str = 'KLDivergence'

# %% ../../nbs/methods/01_vanilla.ipynb 7
class VanillaCF(CFModule):

    def __init__(
        self,
        config: dict | VanillaCFConfig = None,
        name: str = None,
    ):
        if config is None:
            config = VanillaCFConfig()
        config = validate_configs(config, VanillaCFConfig)
        name = "VanillaCF" if name is None else name
        super().__init__(config, name=name)

    def save(self, path: str):
        self.config.save(Path(path) / 'config.json')
    
    @classmethod
    def load_from_path(cls, path: str):
        config = VanillaCFConfig.load_from_json(Path(path) / 'config.json')
        return cls(config=config)

    @auto_reshaping('x')
    def generate_cf(
        self,
        x: Array,  # `x` shape: (k,), where `k` is the number of features
        pred_fn: Callable[[Array], Array],
        y_target: Array = None,
        **kwargs,
    ) -> jnp.DeviceArray:
        # TODO: Currently assumes binary classification.
        if y_target is None:
            y_target = 1 - pred_fn(x)
        else:
            y_target = jnp.array(y_target, copy=True)

        return _vanilla_cf(
            x=x,  # `x` shape: (k,), where `k` is the number of features
            y_target=y_target,  # `y_target` shape: (1,)
            pred_fn=pred_fn,  # y = pred_fn(x)
            n_steps=self.config.n_steps,
            lr=self.config.lr,  # learning rate for each `cf` optimization step
            lambda_=self.config.lambda_,  #  loss = validity_loss + lambda_params * cost
            validity_fn=keras.losses.get({'class_name': self.config.validity_fn, 'config': {'reduction': None}}),
            cost_fn=keras.losses.get({'class_name': 'MeanSquaredError', 'config': {'reduction': None}}),
            apply_constraints_fn=self.apply_constraints,
        )




# AUTOGENERATED! DO NOT EDIT! File to edit: ../nbs/00_base.ipynb.

# %% auto 0
__all__ = ['BaseConfig', 'BaseModule', 'PredFnMixedin', 'TrainableMixedin']

# %% ../nbs/00_base.ipynb 2
from pydantic import BaseModel as BasePydanticModel
import json
from pathlib import Path


# %% ../nbs/00_base.ipynb 4
class BaseConfig(BasePydanticModel):
    """Base class for all config classes."""

    def save(self, path):
        p = Path(path)
        if not str(p).endswith('.json'):
            raise ValueError(f"Path must end with `.json`, but got: {p}")
        if not p.parent.exists():
            p.parent.mkdir(parents=True)
        with open(path, 'w') as f:
            json.dump(self.dict(), f, indent=4)
    
    @classmethod
    def load_from_json(cls, path):
        p = Path(path)
        if not p.exists():
            raise FileNotFoundError(f"File not found: {p}")
        with open(path, 'r') as f:
            return cls(**json.load(f))

# %% ../nbs/00_base.ipynb 6
class BaseModule:
    """Base class for all modules."""
    def __init__(self, config, *, name=None):
        self.config = config
        self._name = name

    @property
    def name(self):
        return self._name or self.__class__.__name__
    
    def save(self, path):
        raise NotImplementedError

    @classmethod
    def load_from_path(cls, path):
        raise NotImplementedError

# %% ../nbs/00_base.ipynb 8
class PredFnMixedin:
    """Mixin class for modules that have a `pred_fn` method."""
    
    def pred_fn(self, x):
        """Return the prediction/probability of the model on `x`."""
        raise NotImplementedError
    
    __ALL__ = ['pred_fn']

# %% ../nbs/00_base.ipynb 9
class TrainableMixedin:
    """Mixin class for trainable modules."""
    
    @property
    def is_trained(self) -> bool:
        """Return whether the module is trained or not."""
        self._is_trained = getattr(self, '_is_trained', False)
        return self._is_trained
    
    def train(self, data, **kwargs):
        """Train the module."""
        raise NotImplementedError
    
    __ALL__ = ['is_trained', 'train']



from .base import CFModule, ParametricCFModule
from .vanilla import VanillaCFConfig, VanillaCF
from .dice import DiverseCF, DiverseCFConfig
from .proto import ProtoCF, ProtoCFConfig
from .counternet import CounterNet, CounterNetConfig
from .sphere import GrowingSphere, GSConfig
from .vaecf import VAECF, VAECFConfig
from .cchvae import CCHVAE, CCHVAEConfig
from .clue import CLUEConfig, CLUE
from .l2c import L2C, L2CConfig



# AUTOGENERATED! DO NOT EDIT! File to edit: ../nbs/02_ml_model.ipynb.

# %% auto 0
__all__ = ['MLPBlock', 'MLP', 'MLModuleConfig', 'MLModule', 'download_ml_module', 'load_ml_module', 'AutoEncoder']

# %% ../nbs/02_ml_model.ipynb 1
from .import_essentials import *
from .data_module import DataModule, DEFAULT_DATA_CONFIGS
from .utils import validate_configs
from .base import *
from sklearn.datasets import make_classification
from sklearn.model_selection import train_test_split
from urllib.request import urlretrieve

# %% ../nbs/02_ml_model.ipynb 3
class MLPBlock(keras.layers.Layer):
    """MLP block with leaky relu activation and dropout/batchnorm."""

    def __init__(
        self, 
        output_size: int, 
        dropout_rate: float = 0.3,
        use_batch_norm: bool = False
    ):
        super().__init__()
        self.output_size = output_size
        self.dropout_rate = dropout_rate
        self.use_batch_norm = use_batch_norm
        if use_batch_norm and dropout_rate > 0:
            warnings.warn("Batch normalization and dropout are usually mutually exclusive.")

    def build(self, input_shape):
        self.dense = keras.layers.Dense(
            self.output_size, activation='leaky_relu', 
            kernel_initializer='he_normal'
        )
        self.dropout = keras.layers.Dropout(self.dropout_rate)
        if self.use_batch_norm:
            self.batch_norm = keras.layers.BatchNormalization()

    def call(self, x, training=False):
        x = self.dense(x)
        x = self.dropout(x, training=training)
        if self.use_batch_norm:
            x = self.batch_norm(x, training=training)
        return x

@keras.saving.register_keras_serializable()
class MLP(keras.Model):
    """MLP model with multiple MLP blocks and a dense layer at the end."""
    
    def __init__(
        self, 
        sizes: list, 
        output_size: int = 2,
        dropout_rate: float = 0.3,
        use_batch_norm: bool = False,
        last_activation: str = 'softmax',
        **kwargs
    ):
        super().__init__(**kwargs)
        self.blocks = []
        for size in sizes:
            self.blocks.append(MLPBlock(size, dropout_rate, use_batch_norm))
        self.dense = keras.layers.Dense(output_size, activation=last_activation)

    def call(self, x, training=False):
        for block in self.blocks:
            x = block(x, training=training)
        return self.dense(x)

    def get_config(self):
        return {
            'sizes': [block.output_size for block in self.blocks],
            'output_size': self.dense.units,
            'dropout_rate': self.blocks[0].dropout_rate,
            'last_activation': self.dense.activation.__name__,
        }

# %% ../nbs/02_ml_model.ipynb 4
class MLModuleConfig(BaseConfig):
    """Configurator of `MLModule`."""
    
    sizes: List[int] = Field([64, 32, 16], description="List of hidden layer sizes.")
    output_size: int = Field(2, description="The number of output classes.")
    dropout_rate: float = Field(0.3, description="Dropout rate.")
    lr: float = Field(1e-3, description="Learning rate.")
    opt_name: str = Field("adam", description="Optimizer name.")
    loss: str = Field("sparse_categorical_crossentropy", description="Loss function name.")
    metrics: List[str] = Field(["accuracy"], description="List of metrics names.")


# %% ../nbs/02_ml_model.ipynb 5
class MLModule(BaseModule, TrainableMixedin, PredFnMixedin):
    def __init__(self, config: MLModuleConfig = None, *, model: keras.Model = None, name: str = None):
        if config is None:
            config = MLModuleConfig()
        config = validate_configs(config, MLModuleConfig)
        self.model = self._init_model(config, model)
        self._is_trained = False
        super().__init__(config, name=name)

    def _init_model(self, config: MLModuleConfig, model: keras.Model):
        if model is None:
            model = MLP(
                sizes=config.sizes,
                output_size=config.output_size,
                dropout_rate=config.dropout_rate
            )
        model.compile(
            optimizer=keras.optimizers.get({
                'class_name': config.opt_name, 
                'config': {'learning_rate': config.lr}
            }),
            loss=config.loss,
            metrics=config.metrics
        )
        return model
            
    def train(
        self, 
        data: DataModule, 
        batch_size: int = 128,
        epochs: int = 10,
        **fit_kwargs
    ):
        if isinstance(data, DataModule):
            X_train, y_train = data['train']
        else:
            X_train, y_train = data
        self.model.fit(
            X_train, y_train, 
            batch_size=batch_size, 
            epochs=epochs,
            **fit_kwargs
        )
        self._is_trained = True
        return self
    
    @property
    def is_trained(self) -> bool:
        return self._is_trained
    
    def save(self, path):
        path = Path(path)
        if not path.exists():
            path.mkdir(parents=True)
        # self.model.save_weights(path / "model.weights.h5", overwrite=True)
        self.model.save(path / "model.keras")
        with open(path / "config.json", "w") as f:
            json.dump(self.config.dict(), f)

    @classmethod
    def load_from_path(cls, path):
        path = Path(path)
        config = MLModuleConfig(**json.load(open(path / "config.json")))
        # model = keras.models.load_model(path / "model.keras")
        model = keras.saving.load_model(path / "model.keras", compile=False)
        module = cls(config, model=model)
        module._is_trained = True
        return module
    
    def pred_fn(self, x):
        if not self.is_trained:
            raise ValueError("Model is not trained.")
        return self.model(x, training=False)

# %% ../nbs/02_ml_model.ipynb 16
def download_ml_module(name: str, path: str = None):
    if path is None:
        path = Path('relax-assets') / name / 'model'
    else:
        path = Path(path)
    if not path.exists():
        path.mkdir(parents=True)
    model_url = f"https://huggingface.co/datasets/birkhoffg/ReLax-Assets/resolve/main/{name}/model/model.keras"
    config_url = f"https://huggingface.co/datasets/birkhoffg/ReLax-Assets/resolve/main/{name}/model/config.json"

    if not (path / "model.keras").exists():
        urlretrieve(model_url, filename=str(path / "model.keras"))
    if not (path / "config.json").exists():
        urlretrieve(config_url, filename=str(path / "config.json"))   
    

def load_ml_module(name: str) -> MLModule:
    """Load the ML module"""

    if name not in DEFAULT_DATA_CONFIGS.keys():
        raise ValueError(f'`data_name` must be one of {DEFAULT_DATA_CONFIGS.keys()}, '
            f'but got data_name={name}.')

    download_ml_module(name)
    return MLModule.load_from_path(f"relax-assets/{name}/model")

# %% ../nbs/02_ml_model.ipynb 19
class AutoEncoder(keras.Model):
    def __init__(
        self,
        enc_sizes: List[int],
        dec_sizes: List[int],
        output_size: int,
        dropout_rate: float = 0.2,
        last_activation: str = "sigmoid",
        name: str = "autoencoder",
        **kwargs
    ):
        super().__init__(name=name, **kwargs)
        self.encoder = keras.Sequential(
            [MLPBlock(size, dropout_rate=dropout_rate) for size in enc_sizes]
        )
        self.decoder = keras.Sequential(
            [MLPBlock(size, dropout_rate=dropout_rate) for size in dec_sizes]
        )
        self.output_layer = keras.layers.Dense(output_size, activation=last_activation)
    
    def call(self, x, training=False):
        z = self.encoder(x, training=training)
        x = self.decoder(z, training=training)
        reconstructed = self.output_layer(x, training=training)
        return reconstructed



# AUTOGENERATED! DO NOT EDIT! File to edit: ../../nbs/methods/02_dice.ipynb.

# %% ../../nbs/methods/02_dice.ipynb 3
from __future__ import annotations
from ..import_essentials import *
from .base import CFModule
from ..base import BaseConfig
from ..utils import auto_reshaping, grad_update, validate_configs

# %% auto 0
__all__ = ['dpp_style_vmap', 'DiverseCFConfig', 'DiverseCF']

# %% ../../nbs/methods/02_dice.ipynb 6
@jit
def dpp_style_vmap(cfs: Array):
    def dpp_fn(cf_1, cf_2):
        return 1 / (1 + jnp.linalg.norm(cf_1 - cf_2, ord=1))
    
    det_entries = vmap(vmap(dpp_fn, in_axes=(None, 0)), in_axes=(0, None))(cfs, cfs)
    det_entries += jnp.eye(cfs.shape[0]) * 1e-8
    assert det_entries.shape == (cfs.shape[0], cfs.shape[0])
    return jnp.linalg.det(det_entries)

# %% ../../nbs/methods/02_dice.ipynb 13
@ft.partial(jit, static_argnums=(2, 3, 4, 8, 9, 10, 11))
def _diverse_cf(
    x: jnp.DeviceArray,  # `x` shape: (k,), where `k` is the number of features
    y_target: Array, # `y_target` shape: (1,)
    pred_fn: Callable[[Array], Array],  # y = pred_fn(x)
    n_cfs: int,
    n_steps: int,
    lr: float,  # learning rate for each `cf` optimization step
    lambdas: Tuple[float, float, float, float], # (lambda_1, lambda_2, lambda_3, lambda_4)
    key: jrand.PRNGKey,
    validity_fn: Callable,
    cost_fn: Callable,
    apply_constraints_fn: Callable,
    compute_reg_loss_fn: Callable,
) -> Array:  # return `cf` shape: (k,)
    """Diverse Counterfactuals (Dice) algorithm."""

    def loss_fn(
        cfs: Array, # shape: (n_cfs, k)
        x: Array, # shape: (1, k)
        pred_fn: Callable[[Array], Array], # y = pred_fn(x)
        y_target: Array,
    ):
        def loss_fn_per_sample(cf, x):
            cf = cf.reshape(1, -1)
            cf_y_pred = pred_fn(cf)
            loss_1 = validity_fn(y_target, cf_y_pred)
            loss_2 = cost_fn(x, cf)
            loss_3 = - dpp_style_vmap(cf)
            loss_4 = compute_reg_loss_fn(x, cf)
            return (
                lambda_1 * loss_1 + 
                lambda_2 * loss_2 + 
                lambda_3 * loss_3 + 
                lambda_4 * loss_4
            )

        return jax.vmap(loss_fn_per_sample, in_axes=(0, None))(cfs, x).mean()
    
    @loop_tqdm(n_steps)
    def gen_cf_step(i, states: Tuple[Array, optax.OptState]):
        cf, opt_state = states
        grads = jax.grad(loss_fn)(cf, x, pred_fn, y_target)
        # grads = jax.grad(loss_fn)(cf, x, pred_fn, y_target)
        cf_updates, opt_state = grad_update(grads, cf, opt_state, opt)
        return cf_updates, opt_state
    
    lambda_1, lambda_2, lambda_3, lambda_4 = lambdas
    key, subkey = jrand.split(key)
    cfs = jrand.uniform(key, (n_cfs, x.shape[-1]))
    opt = optax.adam(lr)
    opt_state = opt.init(cfs)
    
    cfs, opt_state = lax.fori_loop(0, n_steps, gen_cf_step, (cfs, opt_state))
    # TODO: support return multiple cfs
    # cfs = apply_constraints_fn(x, cfs[:1, :], hard=True)
    cfs = apply_constraints_fn(x, cfs, hard=True)
    return cfs


# %% ../../nbs/methods/02_dice.ipynb 15
class DiverseCFConfig(BaseConfig):
    n_cfs: int = 5
    n_steps: int = 1000
    lr: float = 0.001
    lambda_1: float = 1.0
    lambda_2: float = 1.0
    lambda_3: float = 1.0
    lambda_4: float = 0.1
    validity_fn: str = 'KLDivergence'
    cost_fn: str = 'MeanSquaredError'
    seed: int = 42


# %% ../../nbs/methods/02_dice.ipynb 16
class DiverseCF(CFModule):

    def __init__(self, config: dict | DiverseCF = None, *, name: str = None):
        if config is None:
             config = DiverseCFConfig()
        config = validate_configs(config, DiverseCFConfig)
        name = "DiverseCF" if name is None else name
        super().__init__(config, name=name)
    
    def save(self, path: str):
        self.config.save(Path(path) / 'config.json')
    
    @classmethod
    def load_from_path(cls, path: str):
        config = DiverseCFConfig.load_from_json(Path(path) / 'config.json')
        return cls(config=config)

    @auto_reshaping('x', reshape_output=False)
    def generate_cf(
        self,
        x: Array,  # `x` shape: (k,), where `k` is the number of features
        pred_fn: Callable[[Array], Array],
        y_target: Array = None,
        rng_key: jnp.ndarray = None,
        **kwargs,
    ) -> jnp.DeviceArray:
        # TODO: Currently assumes binary classification.
        if y_target is None:
            y_target = 1 - pred_fn(x)
        else:
            y_target = jnp.array(y_target, copy=True).reshape(1, -1)
        if rng_key is None:
            raise ValueError("`rng_key` must be provided.")
        
        assert y_target.shape == (1, 2)
        # print(y_target)
        return _diverse_cf(
            x=x,  # `x` shape: (k,), where `k` is the number of features
            y_target=y_target,  # `y_target` shape: (1,)
            pred_fn=pred_fn,  # y = pred_fn(x)
            n_cfs=self.config.n_cfs,
            n_steps=self.config.n_steps,
            lr=self.config.lr,  # learning rate for each `cf` optimization step
            lambdas=(
                self.config.lambda_1, self.config.lambda_2, 
                self.config.lambda_3, self.config.lambda_4
            ),
            key=rng_key,
            validity_fn=keras.losses.get({'class_name': self.config.validity_fn, 'config': {'reduction': None}}),
            cost_fn=keras.losses.get({'class_name': self.config.cost_fn, 'config': {'reduction': None}}),
            apply_constraints_fn=self.apply_constraints,
            compute_reg_loss_fn=self.compute_reg_loss,
        )

